#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from qtpy.QtCore import Qt
from qtpy.QtCore import Signal
from qtpy.QtWidgets import *


class CameraWidget(QWidget):
    signal_camera_translated = Signal(object)
    signal_camera_rotated = Signal(object)
    signal_center_translated = Signal(object)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle('Camera Properties')
        self.setWindowFlags(Qt.WindowStaysOnTopHint)

        self.label_position = QLabel('Camera position (location)', self)
        self.label_center = QLabel('Rotation center (location)', self)
        self.label_rotation = QLabel('Rotation angle (degrees)', self)

        self.doubleSpinBox_x = self._generate_spinbox()
        self.doubleSpinBox_y = self._generate_spinbox()
        self.doubleSpinBox_z = self._generate_spinbox()

        self.doubleSpinBox_center_x = self._generate_spinbox()
        self.doubleSpinBox_center_y = self._generate_spinbox()
        self.doubleSpinBox_center_z = self._generate_spinbox()

        self.doubleSpinBox_rot_x = self._generate_spinbox(-360, 360)
        self.doubleSpinBox_rot_y = self._generate_spinbox(-360, 360)
        self.doubleSpinBox_rot_z = self._generate_spinbox(-360, 360)

        self.line_separator_1 = QFrame(self)
        self.line_separator_1.setFrameShape(QFrame.HLine)
        self.line_separator_1.setFrameShadow(QFrame.Sunken)

        self.line_separator_2 = QFrame(self)
        self.line_separator_2.setFrameShape(QFrame.HLine)
        self.line_separator_2.setFrameShadow(QFrame.Sunken)

        self.vertical_spacer = QSpacerItem(10, 10, QSizePolicy.Minimum, QSizePolicy.Expanding)

        # Layouts
        self.layout = QVBoxLayout(self)
        self.horizontal_position = QHBoxLayout()
        self.horizontal_center = QHBoxLayout()
        self.horizontal_rotation = QHBoxLayout()

        self.add_to_layout(self.layout, [self.label_position,
                                         self._label_spinbox('X', self.doubleSpinBox_x),
                                         self._label_spinbox('Y', self.doubleSpinBox_y),
                                         self._label_spinbox('Z', self.doubleSpinBox_z),
                                         self.line_separator_1,

                                         self.label_center,
                                         self._label_spinbox('X', self.doubleSpinBox_center_x),
                                         self._label_spinbox('Y', self.doubleSpinBox_center_y),
                                         self._label_spinbox('Z', self.doubleSpinBox_center_z),
                                         self.line_separator_2,

                                         self.label_rotation,
                                         self._label_spinbox('X°', self.doubleSpinBox_rot_x),
                                         self._label_spinbox('Y°', self.doubleSpinBox_rot_y),
                                         self._label_spinbox('Z°', self.doubleSpinBox_rot_z),
                                         ])

        self.layout.addSpacerItem(self.vertical_spacer)

        self._connect_internal_signals()

    @staticmethod
    def add_to_layout(layout, widgets: list) -> None:
        for widget in widgets:
            layout.addWidget(widget)

    def _generate_spinbox(self, lower: float = -9999999.9, upper: float = +9999999.9) -> QAbstractSpinBox:
        spinbox = QDoubleSpinBox(self)
        spinbox.setMinimum(lower)
        spinbox.setMaximum(upper)

        return spinbox

    def _label_spinbox(self, text: str, spinbox: QAbstractSpinBox) -> QWidget:
        widget = QWidget(self)
        layout = QHBoxLayout(widget)
        layout.setContentsMargins(0, 0, 0, 0)

        layout.addWidget(QLabel(text, widget))
        layout.addWidget(spinbox)
        layout.setStretch(0, 1)
        layout.setStretch(1, 8)

        return widget

    def _connect_internal_signals(self) -> None:
        self.doubleSpinBox_x.valueChanged.connect(self.signal_camera_translated.emit)
        self.doubleSpinBox_y.valueChanged.connect(self.signal_camera_translated.emit)
        self.doubleSpinBox_z.valueChanged.connect(self.signal_camera_translated.emit)

        self.doubleSpinBox_rot_x.valueChanged.connect(self.signal_camera_rotated.emit)
        self.doubleSpinBox_rot_y.valueChanged.connect(self.signal_camera_rotated.emit)
        self.doubleSpinBox_rot_z.valueChanged.connect(self.signal_camera_rotated.emit)

        self.doubleSpinBox_center_x.valueChanged.connect(self.signal_center_translated.emit)
        self.doubleSpinBox_center_y.valueChanged.connect(self.signal_center_translated.emit)
        self.doubleSpinBox_center_z.valueChanged.connect(self.signal_center_translated.emit)

    def connect_viewer(self, viewer) -> None:
        # Connect viewer's signals to automatically update self
        viewer.signal_camera_rotated.connect(self.set_rotation_angle)
        viewer.signal_camera_translated.connect(self.set_camera_position)
        viewer.signal_center_translated.connect(self.set_rotation_center)

        # Connect signals to automatically update the viewer
        def angle_setter():
            viewer.rotation_angle = self.get_rotation_angle()

        def camera_setter():
            viewer.camera_position = self.get_camera_position()

        def center_setter():
            viewer.rotation_center = self.get_rotation_center()

        self.signal_camera_rotated.connect(angle_setter)
        self.signal_camera_translated.connect(camera_setter)
        self.signal_center_translated.connect(center_setter)

        # Fill the widget with viewer attributes
        self.set_camera_position(viewer.get_camera_position())
        self.set_rotation_angle(viewer.get_rotation_angle())
        self.set_rotation_center(viewer.get_rotation_center())

    def get_camera_position(self) -> list:
        return [self.doubleSpinBox_x.value(),
                self.doubleSpinBox_y.value(),
                self.doubleSpinBox_z.value()]

    def get_rotation_angle(self) -> list:
        return [self.doubleSpinBox_rot_x.value(),
                self.doubleSpinBox_rot_y.value(),
                self.doubleSpinBox_rot_z.value()]

    def get_rotation_center(self) -> list:
        return [self.doubleSpinBox_center_x.value(),
                self.doubleSpinBox_center_y.value(),
                self.doubleSpinBox_center_z.value()]

    def set_camera_position(self, position: list) -> None:
        self.blockSignals(True)
        self.doubleSpinBox_x.setValue(position[0])
        self.doubleSpinBox_y.setValue(position[1])
        self.doubleSpinBox_z.setValue(position[2])
        self.blockSignals(False)

    def set_rotation_angle(self, angle: list) -> None:
        self.blockSignals(True)
        self.doubleSpinBox_rot_x.setValue(angle[0])
        self.doubleSpinBox_rot_y.setValue(angle[1])
        self.doubleSpinBox_rot_z.setValue(angle[2])
        self.blockSignals(False)

    def set_rotation_center(self, center: list) -> None:
        self.blockSignals(True)
        self.doubleSpinBox_center_x.setValue(center[0])
        self.doubleSpinBox_center_y.setValue(center[1])
        self.doubleSpinBox_center_z.setValue(center[2])
        self.blockSignals(False)



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import numpy as np

from OpenGL.GL import *

from qtpy.QtCore import QFileInfo
from qtpy.QtCore import QPoint
from qtpy.QtCore import QTimer
from qtpy.QtCore import Signal
from qtpy.QtGui import QMatrix4x4
from qtpy.QtGui import QPixmap
from qtpy.QtGui import QQuaternion
from qtpy.QtGui import QRegion
from qtpy.QtGui import QVector3D
from qtpy.QtGui import QVector4D
from qtpy.QtWidgets import QOpenGLWidget
from scipy.spatial.transform import Rotation

from .collections.glaxiscollection import GLAxisCollection
from .collections.glbackgroundcollection import GLBackgroundCollection
from .collections.gldrawablecollection import GLDrawableCollection

from .drawables.gldrawable import GLDrawable
from .drawables.blockgl import BlockGL
from .drawables.linegl import LineGL
from .drawables.meshgl import MeshGL
from .drawables.pointgl import PointGL
from .drawables.tubegl import TubeGL

from .drawablefactory import DrawableFactory
from .fpscounter import FPSCounter

from ..controller.mode import Mode
from ..controller.detectionmode import DetectionMode
from ..controller.normalmode import NormalMode
from ..controller.slicemode import SliceMode
from ..controller.measurementmode import MeasurementMode

from ..model import utils
from ..model.model import Model


class IntegrableViewer(QOpenGLWidget):
    # Signals
    signal_file_modified = Signal()
    signal_load_success = Signal(int)
    signal_load_failure = Signal()
    signal_export_success = Signal(int)
    signal_export_failure = Signal()

    signal_mesh_distances = Signal(object)
    signal_slice_description = Signal(object)
    signal_elements_detected = Signal(object)

    signal_process_updated = Signal(int)
    signal_process_started = Signal()
    signal_process_finished = Signal()

    signal_animation_updated = Signal(int)
    signal_animation_started = Signal()
    signal_animation_finished = Signal()

    signal_screen_clicked = Signal(object)
    signal_mode_updated = Signal(str)
    signal_fps_updated = Signal(float)

    signal_camera_rotated = Signal(object)
    signal_camera_translated = Signal(object)
    signal_center_translated = Signal(object)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setAcceptDrops(True)

        # Model
        self.model = Model()

        # Factory
        self.factory = DrawableFactory(self.model)

        # Drawable elements
        self.axis_collection = GLAxisCollection(self)
        self.background_collection = GLBackgroundCollection(self)
        self.drawable_collection = GLDrawableCollection(self)

        # Controllers
        self.current_mode = None
        self.controllers = {}

        # Camera/World/Projection
        self.camera = QMatrix4x4()
        self.world = QMatrix4x4()
        self.proj = QMatrix4x4()

        # FPS Counter
        self.fps_counter = FPSCounter()

        # Initial positions and rotations
        self.rotation_center = [0.0, 0.0, 0.0]
        self.rotation_angle = [0.0, 0.0, 0.0]
        self.camera_position = [0.0, 0.0, 200.0]

        # Extra information
        self._turbo = False
        self._autofit = False
        self._animated = False

        self.fov = 45.0
        self.smoothness = 2.0  # Bigger => smoother (but slower) rotations
        self.projection_mode = 'perspective'  # 'perspective'/'orthographic'

        # Initialize viewer
        self.initialize()

    def initialize(self) -> None:
        # Axis/Background
        self.register_drawable(self.factory.axis(), self.axis_collection)
        self.register_drawable(self.factory.background(), self.background_collection)

        # Signals
        self.signal_file_modified.connect(self.recreate)
        self.signal_camera_rotated.connect(self.update)
        self.signal_camera_translated.connect(self.update)
        self.signal_center_translated.connect(self.update)

        # Controllers
        self.add_controller(NormalMode(self))
        self.add_controller(DetectionMode(self))
        self.add_controller(SliceMode(self))
        self.add_controller(MeasurementMode(self))
        self.set_normal_mode()

        # FPSCounter
        self.fps_counter.add_callback(self.signal_fps_updated.emit)

    """
    Properties
    """
    @property
    def last_id(self) -> int:
        return self.drawable_collection.last_id

    @property
    def last_drawable(self) -> GLDrawable:
        return self.get_drawable(self.last_id)

    @property
    def axis(self) -> GLDrawable:
        return self.axis_collection.get_last()

    @property
    def background(self) -> GLDrawable:
        return self.background_collection.get_last()

    @property
    def off_center(self) -> np.ndarray:
        return self.camera_position - self.rotation_center

    """
    Camera movements/rotations
    """
    @property
    def camera_position(self) -> np.ndarray:
        return np.array([self.xCameraPos, self.yCameraPos, self.zCameraPos])

    @property
    def rotation_angle(self) -> np.ndarray:
        return np.array([self.xCenterRot, self.yCenterRot, self.zCenterRot])

    @property
    def rotation_center(self) -> np.ndarray:
        return np.array([self.xCenterPos, self.yCenterPos, self.zCenterPos])

    @camera_position.setter
    def camera_position(self, pos: list) -> None:
        self.xCameraPos, self.yCameraPos, self.zCameraPos = pos
        self.signal_camera_translated.emit(pos)

    @rotation_angle.setter
    def rotation_angle(self, rot: list) -> None:
        self.xCenterRot, self.yCenterRot, self.zCenterRot = rot
        self.signal_camera_rotated.emit(rot)

    @rotation_center.setter
    def rotation_center(self, center: list) -> None:
        self.xCenterPos, self.yCenterPos, self.zCenterPos = center
        self.signal_center_translated.emit(center)

    """
    API for partial camera movements/rotations
    """
    def translate(self, x: float, y: float, z: float) -> None:
        self.camera_position += np.array([x, y, z])

    def translate_center(self, x: float, y: float, z: float) -> None:
        self.rotation_center += np.array([x, y, z])

    def rotate(self, alpha: float, beta: float, gamma: float) -> None:
        self.rotation_angle += np.array([alpha, beta, gamma])
        self.rotation_angle %= 360

    """
    API for the camera (animated)
    """
    def get_camera_position(self) -> np.ndarray:
        return self.camera_position

    def get_rotation_center(self) -> np.ndarray:
        return self.rotation_center

    def get_rotation_angle(self) -> np.ndarray:
        return self.rotation_angle

    def set_camera_position(self, target) -> None:
        origin = self.get_camera_position()

        def setter(value):
            self.camera_position = value
        self.animate(origin, target, method=setter) if self.is_animated() else setter(target)

    def set_rotation_center(self, target) -> None:
        origin = self.get_rotation_center()

        def setter(value):
            self.rotation_center = value
        self.animate(origin, target, method=setter) if self.is_animated() else setter(target)

    def set_rotation_angle(self, target) -> None:
        # Fix any difference > 180° so the camera doesn't over-rotate
        origin = self.get_rotation_angle() % 360
        origin[origin - target > 180] -= 360
        origin[origin - target < -180] += 360

        def setter(value):
            self.rotation_angle = value
        self.animate(origin, target, method=setter) if self.is_animated() else setter(target)

    """
    Animation / Turbo
    """
    def set_turbo_rendering(self, status: bool) -> None:
        self.blockSignals(True)

        for d in self.get_all_drawables():
            d.is_boostable = status

        self.blockSignals(False)
        self.recreate()

    def is_animated(self) -> bool:
        return self._animated

    def set_animated(self, status: bool) -> None:
        self._animated = status

    def animate(self, start, end, method: callable, **kwargs) -> callable:
        milliseconds = int(kwargs.get('milliseconds', 300))
        steps = int(kwargs.get('steps', 20))

        timer = QTimer(self)
        linspace = iter(np.linspace(start, end, steps))
        counter = []

        def start_animation():
            timer.start(milliseconds / steps)
            self.signal_animation_started.emit()

        def update_animation():
            try:
                method(next(linspace))
                self.update()

                counter.append(None)
                self.signal_animation_updated.emit(100.0 * len(counter) / steps)
            except StopIteration:
                end_animation()

        def end_animation():
            timer.stop()
            self.signal_animation_finished.emit()

        timer.timeout.connect(update_animation)
        start_animation()

    """
    Projections
    """
    def perspective_projection(self) -> None:
        self.projection_mode = 'perspective'
        self.update()

    def orthographic_projection(self) -> None:
        self.projection_mode = 'orthographic'
        self.update()

    """
    Registration methods
    """
    def register_drawable(self, drawable, collection=None):
        if drawable is None:
            self.signal_load_failure.emit()
            return None

        if collection is None:
            collection = self.drawable_collection

        # Register in collection
        drawable.add_observer(self)
        collection.add(drawable)

        # Emit success signals
        self.signal_file_modified.emit()
        self.signal_load_success.emit(drawable.id)

        return drawable

    """
    Load methods by arguments
    """
    def mesh(self, *args, **kwargs) -> MeshGL:
        return self.register_drawable(self.factory.mesh(*args, **kwargs))

    def blocks(self, *args, **kwargs) -> BlockGL:
        return self.register_drawable(self.factory.blocks(*args, **kwargs))

    def points(self, *args, **kwargs) -> PointGL:
        return self.register_drawable(self.factory.points(*args, **kwargs))

    def lines(self, *args, **kwargs) -> LineGL:
        return self.register_drawable(self.factory.lines(*args, **kwargs))

    def tubes(self, *args, **kwargs) -> TubeGL:
        return self.register_drawable(self.factory.tubes(*args, **kwargs))

    """
    Load methods by path
    """
    def load_mesh(self, path: str, *args, **kwargs) -> MeshGL:
        return self.register_drawable(self.factory.load_mesh(path, *args, **kwargs))

    def load_blocks(self, path: str, *args, **kwargs) -> BlockGL:
        return self.register_drawable(self.factory.load_blocks(path, *args, **kwargs))

    def load_points(self, path: str, *args, **kwargs) -> PointGL:
        return self.register_drawable(self.factory.load_points(path, *args, **kwargs))

    def load_lines(self, path: str, *args, **kwargs) -> LineGL:
        return self.register_drawable(self.factory.load_lines(path, *args, **kwargs))

    def load_tubes(self, path: str, *args, **kwargs) -> TubeGL:
        return self.register_drawable(self.factory.load_tubes(path, *args, **kwargs))

    def load_mesh_folder(self, path: str, *args, **kwargs) -> list:
        return self.load_folder(path, self.load_mesh, *args, **kwargs)

    def load_blocks_folder(self, path: str, *args, **kwargs) -> list:
        return self.load_folder(path, self.load_blocks, *args, **kwargs)

    def load_points_folder(self, path: str, *args, **kwargs) -> list:
        return self.load_folder(path, self.load_points, *args, **kwargs)

    def load_lines_folder(self, path: str, *args, **kwargs) -> list:
        return self.load_folder(path, self.load_lines, *args, **kwargs)

    def load_tubes_folder(self, path: str, *args, **kwargs) -> list:
        return self.load_folder(path, self.load_tubes, *args, **kwargs)

    """
    Generalized loaders
    """
    def load_folder(self, path: str, loader: callable, *args, **kwargs) -> list:
        return self.load_multiple(self.model.get_paths_from_directory(path), loader, *args, **kwargs)

    def load_multiple(self, path_list: list, loader: callable, *args, **kwargs) -> list:
        # Do not process if loading an empty path list
        if len(path_list) == 0:
            return []

        self.signal_process_started.emit()
        loaded = []

        for i, path in enumerate(path_list, 1):
            # Block only loaded signals
            self.blockSignals(True)
            drawable = loader(path, *args, **kwargs)
            self.blockSignals(False)

            if drawable is not None:
                loaded.append(drawable)

            self.signal_process_updated.emit(int(100 * i / len(path_list)))

        # Emit signals depending on results
        if len(loaded) > 0:
            self.signal_file_modified.emit()
            self.signal_load_success.emit(self.last_id)
        else:
            self.signal_load_failure.emit()

        self.signal_process_finished.emit()

        return loaded

    """
    Export methods
    """
    def export_element(self, path: str, _id: int) -> None:
        try:
            self.model.export(path, _id)
            self.signal_export_success.emit(_id)
        except Exception:
            self.signal_export_failure.emit()

    """
    Drawable manipulation
    """
    def get_drawable(self, _id: int):
        return self.drawable_collection.get(_id)

    def show_drawable(self, _id: int) -> None:
        self.get_drawable(_id).show()

    def hide_drawable(self, _id: int) -> None:
        self.get_drawable(_id).hide()

    def update_drawable(self, _id: int) -> None:
        self.makeCurrent()
        self.get_drawable(_id).setup_attributes()
        self.recreate()

    def delete(self, _id: int) -> None:
        if _id < 0:
            return

        self.makeCurrent()
        self.model.delete(_id)
        self.drawable_collection.delete(_id)
        self.signal_file_modified.emit()

    def update_all(self) -> None:
        self.makeCurrent()
        for _id in self.get_all_ids():
            self.get_drawable(_id).setup_attributes()
        self.recreate()

    def clear(self) -> None:
        self.makeCurrent()
        for _id in self.get_all_ids():
            self.model.delete(_id)
            self.drawable_collection.delete(_id)
        self.signal_file_modified.emit()

    """
    Drawable retrieval
    """
    def get_all_ids(self) -> list:
        return self.drawable_collection.get_all_ids()

    def get_all_drawables(self) -> list:
        return self.drawable_collection.get_all_drawables()

    def get_all_meshes(self) -> list:
        return self.drawable_collection.filter(MeshGL)

    def get_all_blocks(self) -> list:
        return self.drawable_collection.filter(BlockGL)

    def get_all_points(self) -> list:
        return self.drawable_collection.filter(PointGL)

    def get_all_lines(self) -> list:
        return self.drawable_collection.filter(LineGL)

    def get_all_tubes(self) -> list:
        return self.drawable_collection.filter(TubeGL)

    """
    Camera handling
    """
    def camera_at(self, _id: int) -> None:
        self.fit_to_bounds(*self.get_drawable(_id).bounding_box)

    def plan_view(self) -> None:
        self.set_rotation_angle([0.0, 0.0, 0.0])

    def north_view(self) -> None:
        self.set_rotation_angle([270.0, 0.0, 270.0])

    def east_view(self) -> None:
        self.set_rotation_angle([270.0, 0.0, 0.0])

    def fit_to_bounds(self, min_bound: np.ndarray, max_bound: np.ndarray) -> None:
        center = (min_bound + max_bound) / 2
        self.rotation_center = center

        # Put the camera in a position that allow us to see between the boundaries.
        md = np.max(np.diff([min_bound, max_bound], axis=0))
        aspect = self.width() / self.height()
        fov_rad = self.fov * np.pi / 180.0

        # In perspective projection, we need a clever trigonometric calculation.
        if self.projection_mode == 'perspective':
            dist = (md / np.tan(fov_rad / 2) + md) / 2.0
            z_shift = dist * max(1.0, 1.0 / aspect)

        # But in orthographic projection, it's more direct.
        else:
            dist = md / 2
            z_shift = dist * max(1.0, aspect)

        camera_shift = center + np.array([0.0, 0.0, 1.1 * z_shift])
        self.set_camera_position(camera_shift)
        self.update()

    def fit_to_screen(self) -> None:
        bounds = self.bounding_box()

        if bounds is not None:
            self.fit_to_bounds(*bounds)

    def bounding_box(self) -> tuple or None:
        drawables = [d for d in self.get_all_drawables() if d.is_visible]

        if len(drawables) == 0:
            return None

        min_all = np.inf * np.ones(3)
        max_all = -np.inf * np.ones(3)

        for drawable in drawables:
            min_bound, max_bound = drawable.bounding_box
            min_all = np.min((min_all, min_bound), axis=0)
            max_all = np.max((max_all, max_bound), axis=0)

        return min_all, max_all

    """
    Internal methods
    """
    def initializeGL(self) -> None:
        glClearColor(0.0, 0.0, 0.0, 1.0)
        glEnable(GL_VERTEX_PROGRAM_POINT_SIZE)
        glEnable(GL_POINT_SPRITE)
        glDisable(GL_CULL_FACE)
        glEnable(GL_DEPTH_TEST)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

        # Initialize collections
        self.background_collection.initialize()
        self.drawable_collection.initialize()
        self.axis_collection.initialize()

    def paintGL(self) -> None:
        # Clear screen
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

        self.world.setToIdentity()
        self.camera.setToIdentity()
        self.proj.setToIdentity()

        # Translate by rotation center (world position)
        self.world.translate(*self.rotation_center)

        # Allow rotation of the world
        self.world.rotate(self.xCenterRot, 1.0, 0.0, 0.0)
        self.world.rotate(self.yCenterRot, 0.0, 1.0, 0.0)
        self.world.rotate(self.zCenterRot, 0.0, 0.0, 1.0)

        # Restore world
        self.world.translate(*-self.rotation_center)

        # Translate the camera
        self.camera.translate(*-self.camera_position)

        # Project (Perspective/Orthographic)
        self.resizeGL(self.width(), self.height())

        # Update matrices
        self.drawable_collection.update_matrix('proj_matrix', self.proj)
        self.drawable_collection.update_matrix('model_view_matrix', self.camera * self.world)
        self.axis_collection.update_matrix('model_view_matrix', self.camera * self.world)

        # Draw every GLDrawable (meshes, blocks, points, etc)
        glEnable(GL_BLEND)
        self.background_collection.draw()
        self.drawable_collection.draw()
        self.axis_collection.draw()

        # Tick FPS counter
        self.fps_counter.tick()

    def resizeGL(self, w: float, h: float) -> None:
        aspect = w / h

        if self.projection_mode == 'perspective':
            self.proj.perspective(self.fov, aspect, 1.0, 100000.0)
        else:  # if self.projection_mode == 'orthographic':
            z = self.off_center[2]
            self.proj.ortho(-z, z, -z / aspect, z / aspect, 0.0, 100000.0)

    def recreate(self) -> None:
        self.axis_collection.recreate()
        self.background_collection.recreate()
        self.drawable_collection.recreate()
        self.update()

    """
    Utilities
    """
    def get_pixmap(self) -> QPixmap:
        pixmap = QPixmap(self.size())
        self.render(pixmap, QPoint(), QRegion(self.rect()))
        return pixmap

    def take_screenshot(self, save_path: str) -> None:
        if bool(save_path):
            self.get_pixmap().save(save_path)

    def screen_to_ndc(self, x: float, y: float, z: float) -> np.ndarray:
        # Click at bottom-left of screen => (-1.0, -1.0, z)
        # Click at top-right of screen => (1.0, 1.0, z)
        # But we can't really know where's z, so we just return 1.0
        return np.array([
            (2.0 * x / self.width()) - 1.0,
            1.0 - (2.0 * y / self.height()),
            1.0])

    def unproject(self, x: float, y: float, z: float) -> np.ndarray:
        # Adapted from http://antongerdelan.net/opengl/raycasting.html
        # Note: x, y, z must be normalized to [-1.0, +1.0]

        # We'd use `QVector4D(x, y, -1.0, 1.0)`, but PySide2
        # hasn't implemented QMatrix4x4 * QVector4D yet.
        vector = [x, y, -1.0, 1.0]
        temp_matrix = QMatrix4x4(*[e for e in vector for _ in range(4)])

        ray_eye = (self.proj.inverted()[0] * temp_matrix).column(0)
        ray_eye = QVector4D(ray_eye.x(), ray_eye.y(), -1.0, 0.0)

        # We'd use `ray_eye`, but PySide2
        # hasn't implemented QMatrix4x4 * QVector4D yet.
        vector = [ray_eye.x(), ray_eye.y(), ray_eye.z(), ray_eye.w()]
        temp_matrix = QMatrix4x4(*[e for e in vector for _ in range(4)])

        ray_world = ((self.camera * self.world).inverted()[0] * temp_matrix).column(0).toVector3D()
        ray = ray_world.normalized()
        return np.array([ray.x(), ray.y(), ray.z()])

    def ray_from_click(self, x: float, y: float, z: float) -> np.ndarray:
        # Perspective projection is straightforward
        def perspective_ray() -> np.ndarray:
            return self.unproject(*self.screen_to_ndc(x, y, z))

        # Orthographic projection "forces" a click in the center (normalized to [-1.0, +1.0])
        def orthographic_ray() -> np.ndarray:
            return self.unproject(0.0, 0.0, 0.0)

        if self.projection_mode == 'perspective':
            return perspective_ray()

        return orthographic_ray()

    def origin_from_click(self, x: float, y: float, z: float) -> np.ndarray:
        # Perspective projection is straightforward
        def perspective_origin() -> np.ndarray:
            origin = (self.camera * self.world).inverted()[0].column(3).toVector3D()
            return np.array([origin.x(), origin.y(), origin.z()])

        # Orthographic projection needs a bit more of vector arithmetic.
        # A click in the center of the screen gives us the perfect ray,
        # by recycling the perspective's un-project method.

        # But if we don't click in the exact center of screen,
        # we need to trick the origin calculation.
        def orthographic_origin() -> np.ndarray:
            ndc = self.screen_to_ndc(x, y, z)
            aspect = self.width() / self.height()
            aspect_bias = np.array([1.0, 1.0 / aspect, 0.0])

            # The idea is to strategically move the camera by an offset, so that
            # clicking anywhere in the screen will give us the same result as
            # clicking at the exact center of the screen from a shifted camera.
            offset = self.off_center[2] * ndc * aspect_bias

            # Hack to get the origin when the click is not exactly at center of screen.
            self.camera.translate(*-offset)
            origin = perspective_origin()
            self.camera.translate(*offset)

            return origin

        if self.projection_mode == 'perspective':
            return perspective_origin()

        return orthographic_origin()

    def get_normal(self, origin_list, ray_list: list) -> np.ndarray:
        # Perspective: Same origins, different rays
        def perspective_normal() -> np.ndarray:
            return utils.normalize(np.cross(*ray_list))

        def orthographic_normal() -> np.ndarray:
            # Orthographic: Same rays, different origins
            origin_diff = np.diff(origin_list, axis=0)[0]
            return utils.normalize(np.cross(ray_list[0], origin_diff))

        if self.projection_mode == 'perspective':
            return perspective_normal()

        return orthographic_normal()

    @staticmethod
    def angles_from_vectors(normal: np.ndarray, up: np.ndarray) -> np.ndarray:
        # Returns a list of angles that allows the normal to look directly at the camera
        # Extrinsic: xyz / Intrinsic: XYZ (Tait-Bryan angles)
        # In self.paintGL, we're currently using the XYZ intrinsic rotation order.

        rotmat = QQuaternion.fromDirection(QVector3D(*normal), QVector3D(*up)).toRotationMatrix().data()
        return Rotation.from_matrix(np.array(rotmat).reshape((3, 3))).as_euler('XYZ', degrees=True)

    def set_camera_from_vectors(self, normal: np.ndarray, up: np.ndarray) -> None:
        # Auto-moves the camera using the normal as direction
        self.set_rotation_angle(self.angles_from_vectors(normal, up))

    def generate_slice_description(self, origin_list: list, ray_list: list) -> None:
        # A plane is created from `origin` and `ray_list`.
        # In perspective projection, the origin is the same.
        origin = origin_list[0]
        normal = self.get_normal(origin_list, ray_list)
        up = utils.normalize(np.cross(ray_list[0], normal))

        # Emit description of the slice
        self.signal_slice_description.emit({
            'origin': origin,
            'normal': normal,
            'up': up,
        })

    def slice_meshes(self, origin: np.ndarray, normal: np.ndarray, include_hidden: bool = False) -> list:
        # By default, slice only visible meshes
        meshes = list(filter(lambda m: m.is_visible or include_hidden, self.get_all_meshes()))

        return self.model.slice_meshes(origin, normal, meshes)

    def slice_blocks(self, origin: np.ndarray, normal: np.ndarray, include_hidden: bool = True) -> list:
        # By default, slice visible and hidden blocks
        blocks = list(filter(lambda m: m.is_visible or include_hidden, self.get_all_blocks()))

        return self.model.slice_blocks(origin, normal, blocks)

    def intersect_meshes(self, origin: np.ndarray, ray: np.ndarray, include_hidden: bool = False) -> list:
        # By default, intersect only visible meshes
        meshes = list(filter(lambda m: m.is_visible or include_hidden, self.get_all_meshes()))

        return self.model.intersect_meshes(origin, ray, meshes)

    def intersect_lines(self, origin: np.ndarray, ray: np.ndarray, include_hidden: bool = False) -> list:
        # By default, intersect only visible lines
        lines = list(filter(lambda m: m.is_visible or include_hidden, self.get_all_lines()))

        return self.model.intersect_lines(origin, ray, lines)

    def intersect_elements(self, origin: np.ndarray, ray: np.ndarray, include_hidden: bool = False) -> list:
        results = []
        results += self.intersect_meshes(origin, ray, include_hidden)
        results += self.intersect_lines(origin, ray, include_hidden)

        self.signal_elements_detected.emit(results)

    def measure_from_rays(self, origin_list: list, ray_list: list) -> None:
        meshes = list(filter(lambda m: m.is_visible, self.get_all_meshes()))

        results = self.model.measure_from_rays(origin_list, ray_list, meshes)
        self.signal_mesh_distances.emit(results)

    def cross_section(self, origin: np.ndarray, normal: np.ndarray) -> None:
        self.drawable_collection.update_uniform('CrossSection', 'plane_origin', *origin)
        self.drawable_collection.update_uniform('CrossSection', 'plane_normal', *normal)

    def set_cross_section(self, status: bool) -> None:
        for m in self.get_all_meshes():
            m.is_cross_sectionable = status

    """
    Controller
    """
    def add_controller(self, controller: Mode) -> None:
        self.controllers[controller.name] = controller

    def set_controller(self, mode_name: str) -> None:
        self.current_mode = self.controllers.get(mode_name)
        self.signal_mode_updated.emit(self.current_mode.name)
        self.update()

    def set_normal_mode(self) -> None:
        self.set_controller('Normal Mode')

    def set_detection_mode(self) -> None:
        self.set_controller('Detection Mode')

    def set_slice_mode(self) -> None:
        self.set_controller('Slice Mode')

    def set_measurement_mode(self) -> None:
        self.set_controller('Measurement Mode')

    """
    Events (dependent on current controller)
    """
    def mouseMoveEvent(self, event, *args, **kwargs) -> None:
        self.current_mode.mouseMoveEvent(event)
        self.update()

    def mousePressEvent(self, event, *args, **kwargs) -> None:
        self.current_mode.mousePressEvent(event)
        self.update()

    def mouseDoubleClickEvent(self, event, *args, **kwargs) -> None:
        self.current_mode.mouseDoubleClickEvent(event)
        self.update()

    def mouseReleaseEvent(self, event, *args, **kwargs) -> None:
        self.current_mode.mouseReleaseEvent(event)
        self.update()

    def wheelEvent(self, event, *args, **kwargs) -> None:
        self.current_mode.wheelEvent(event)
        self.update()

    def dragEnterEvent(self, event, *args, **kwargs) -> None:
        if event.mimeData().hasText():
            event.acceptProposedAction()

    def dropEvent(self, event, *args, **kwargs) -> None:
        for url in event.mimeData().urls():
            path = url.toLocalFile()
            if QFileInfo(path).isDir():
                self.load_folder(path, self.load_mesh)
            else:
                self.load_mesh(path)



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import numpy as np

from .gldrawable import GLDrawable
from OpenGL.GL import *


class LineGL(GLDrawable):
    def __init__(self, element, *args, **kwargs):
        super().__init__(element, *args, **kwargs)
        self.num_vertices = 0

    def setup_attributes(self) -> None:
        _POSITION = 0
        _COLOR = 1

        # Generate VAO and VBOs (see GLDrawable)
        self.create_vao_vbos(2)

        # Data
        vertices = self.element.vertices.astype(np.float32)
        colors = self.element.rgba.astype(np.float32)

        # np.array([[0, 1, 2]], type) has size 3, despite having only 1 list there
        self.num_vertices = len(vertices)

        glBindVertexArray(self.vao)

        # Fill buffers (see GLDrawable)
        self.fill_buffer(_POSITION, 3, vertices, GLfloat, GL_FLOAT, self.vbos[_POSITION])
        self.fill_buffer(_COLOR, 4, colors, GLfloat, GL_FLOAT, self.vbos[_COLOR])

        glVertexAttribDivisor(_COLOR, 1)

        glBindVertexArray(0)

    def draw(self) -> None:
        glBindVertexArray(self.vao)
        glLineWidth(self.element.thickness)
        glDrawArrays(GL_LINE_STRIP, 0, self.num_vertices)
        glLineWidth(1)
        glBindVertexArray(0)



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from qtpy.QtGui import QMouseEvent
from .mode import Mode


class DetectionMode(Mode):
    def __init__(self, viewer):
        super().__init__(viewer)
        self.name = 'Detection Mode'

    def mousePressEvent(self, event: QMouseEvent) -> None:
        super().mousePressEvent(event)

        x, y, z = [event.pos().x(), event.pos().y(), 1]
        origin = self.viewer.origin_from_click(x, y, z)
        ray = self.viewer.ray_from_click(x, y, z)

        self.viewer.intersect_elements(origin, ray)



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import time


class FPSCounter:
    def __init__(self, resolution: float = 1.0):
        self.start_time = time.time()
        self.resolution = resolution
        self.callbacks = []
        self.counter = 0

    def add_callback(self, callback: callable) -> None:
        self.callbacks.append(callback)

    def pop_callback(self) -> callable:
        return self.callbacks.pop()

    def clear_callbacks(self) -> None:
        self.callbacks.clear()

    def notify(self, fps: float) -> None:
        for callback in self.callbacks:
            callback(fps)

    def tick(self) -> None:
        self.counter += 1
        if time.time() - self.start_time > self.resolution:
            diff = (time.time() - self.start_time)
            self.notify(self.counter / diff if diff else 0.0)

            self.counter = 0
            self.start_time = time.time()



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from .shaderprogram import ShaderProgram


class BackgroundProgram(ShaderProgram):
    def __init__(self, viewer):
        super().__init__(viewer)
        self.base_name = 'Background'

    def initialize(self) -> None:
        super().initialize()
        self.add_uniform_loc('top_color')
        self.add_uniform_loc('bottom_color')

    def inner_draw(self, drawables: list) -> None:
        for drawable in drawables:
            self.update_uniform('top_color', *drawable.top_color)
            self.update_uniform('bottom_color', *drawable.bottom_color)
            drawable.draw()



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from .parser import Parser


class ParserCollection:
    def __init__(self):
        self._collection = {}

    def add(self, extension: str, handler: Parser) -> None:
        self._collection[extension] = handler

    def get(self, extension: str) -> Parser:
        return self._collection.get(extension)

    def delete(self, extension: str) -> None:
        del self._collection[extension]



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from qtpy.QtWidgets import QAction

from .iconcollection import IconCollection


class ActionCollection:
    def __init__(self, parent):
        # Tree Actions
        self.action_show = QAction('&Show', parent)
        self.action_show.setIcon(IconCollection.get('flash_on.svg'))

        self.action_hide = QAction('&Hide', parent)
        self.action_hide.setIcon(IconCollection.get('flash_off.svg'))

        self.action_delete = QAction('&Delete', parent)
        self.action_delete.setIcon(IconCollection.get('cancel.svg'))

        self.action_focus_camera = QAction('&Focus camera', parent)
        self.action_focus_camera.setIcon(IconCollection.get('collect.svg'))

        self.action_highlight = QAction('H&ighlight', parent)
        self.action_highlight.setIcon(IconCollection.get('idea.svg'))
        self.action_highlight.setCheckable(True)

        self.action_wireframe = QAction('&Wireframe', parent)
        self.action_wireframe.setIcon(IconCollection.get('grid.svg'))
        self.action_wireframe.setCheckable(True)

        self.action_properties = QAction('&Properties', parent)
        self.action_properties.setIcon(IconCollection.get('settings.svg'))

        self.action_setup_colors = QAction('Setup c&olors', parent)
        self.action_setup_colors.setIcon(IconCollection.get('picture.svg'))

        self.action_setup_multiple = QAction('Setup &multiple', parent)
        self.action_setup_multiple.setIcon(IconCollection.get('picture.svg'))

        self.action_export_element = QAction('&Export', parent)
        self.action_export_element.setIcon(IconCollection.get('export.svg'))

        # MainWindow Actions
        self.action_load_mesh = QAction('Load &mesh', parent)
        self.action_load_mesh.setIcon(IconCollection.get('area_chart.svg'))
        self.action_load_mesh.setShortcut('Ctrl+A')

        self.action_load_blocks = QAction('Load &blocks', parent)
        self.action_load_blocks.setIcon(IconCollection.get('data_sheet.svg'))
        self.action_load_blocks.setShortcut('Ctrl+E')

        self.action_load_points = QAction('Load &points', parent)
        self.action_load_points.setIcon(IconCollection.get('scatter_plot.svg'))
        self.action_load_points.setShortcut('Ctrl+P')

        self.action_load_lines = QAction('Load &lines', parent)
        self.action_load_lines.setIcon(IconCollection.get('line_chart.svg'))
        self.action_load_lines.setShortcut('Ctrl+L')

        self.action_load_tubes = QAction('Load &tubes', parent)
        self.action_load_tubes.setIcon(IconCollection.get('radar_plot.svg'))
        self.action_load_tubes.setShortcut('Ctrl+T')

        self.action_load_mesh_folder = QAction('Load m&esh folder', parent)
        self.action_load_mesh_folder.setIcon(IconCollection.get('area_chart.svg'))
        self.action_load_mesh_folder.setShortcut('Ctrl+Shift+A')

        self.action_load_blocks_folder = QAction('Load b&locks folder', parent)
        self.action_load_blocks_folder.setIcon(IconCollection.get('data_sheet.svg'))
        self.action_load_blocks_folder.setShortcut('Ctrl+Shift+E')

        self.action_load_points_folder = QAction('Load p&oints folder', parent)
        self.action_load_points_folder.setIcon(IconCollection.get('scatter_plot.svg'))
        self.action_load_points_folder.setShortcut('Ctrl+Shift+P')

        self.action_load_lines_folder = QAction('Load l&ines folder', parent)
        self.action_load_lines_folder.setIcon(IconCollection.get('line_chart.svg'))
        self.action_load_lines_folder.setShortcut('Ctrl+Shift+L')

        self.action_load_tubes_folder = QAction('Load t&ubes folder', parent)
        self.action_load_tubes_folder.setIcon(IconCollection.get('radar_plot.svg'))
        self.action_load_tubes_folder.setShortcut('Ctrl+Shift+T')

        self.action_quit = QAction('&Quit', parent)
        self.action_quit.setIcon(IconCollection.get('cancel.svg'))
        self.action_quit.setShortcut('Ctrl+Q')

        self.action_show_tree = QAction('Show &tree', parent)
        self.action_show_tree.setIcon(IconCollection.get('list.svg'))
        self.action_show_tree.setShortcut('Ctrl+W')

        self.action_help = QAction('&Help', parent)
        self.action_help.setIcon(IconCollection.get('document.svg'))
        self.action_help.setShortcut('F1')

        self.action_about = QAction('&About', parent)
        self.action_about.setIcon(IconCollection.get('about.svg'))
        self.action_about.setShortcut('?')

        self.action_animate = QAction('A&nimate movements', parent)
        self.action_animate.setIcon(IconCollection.get('film_reel.svg'))
        self.action_animate.setShortcut('Shift+N')
        self.action_animate.setCheckable(True)

        self.action_autofit = QAction('&Auto-fit to screen', parent)
        self.action_autofit.setIcon(IconCollection.get('globe.svg'))
        self.action_autofit.setShortcut('Shift+F')
        self.action_autofit.setCheckable(True)

        self.action_autorotate = QAction('Auto-&rotate on slice', parent)
        self.action_autorotate.setIcon(IconCollection.get('process.svg'))
        self.action_autorotate.setShortcut('Shift+R')
        self.action_autorotate.setCheckable(True)

        self.action_turbo_rendering = QAction('&Turbo rendering', parent)
        self.action_turbo_rendering.setIcon(IconCollection.get('sports_mode.svg'))
        self.action_turbo_rendering.setShortcut('Shift+T')
        self.action_turbo_rendering.setCheckable(True)

        self.action_normal_mode = QAction('Return to &normal', parent)
        self.action_normal_mode.setIcon(IconCollection.get('rotate_camera.svg'))
        self.action_normal_mode.setShortcut('Ctrl+N')

        self.action_measurement_mode = QAction('&Measure meshes', parent)
        self.action_measurement_mode.setIcon(IconCollection.get('ruler.svg'))
        self.action_measurement_mode.setShortcut('Ctrl+M')

        self.action_detection_mode = QAction('&Detect meshes/lines', parent)
        self.action_detection_mode.setIcon(IconCollection.get('cursor.svg'))
        self.action_detection_mode.setShortcut('Ctrl+D')

        self.action_slice_meshes = QAction('&Slice meshes', parent)
        self.action_slice_meshes.setIcon(IconCollection.get('flash_on.svg'))
        self.action_slice_meshes.setShortcut('Ctrl+S')

        self.action_slice_blocks = QAction('Slice &blocks', parent)
        self.action_slice_blocks.setIcon(IconCollection.get('flash_on.svg'))
        self.action_slice_blocks.setShortcut('Ctrl+B')

        self.action_cross_section = QAction('&Cross-section', parent)
        self.action_cross_section.setIcon(IconCollection.get('flash_on.svg'))
        self.action_cross_section.setShortcut('Ctrl+Shift+C')
        self.action_cross_section.setCheckable(True)

        self.action_camera_properties = QAction('&Camera properties', parent)
        self.action_camera_properties.setIcon(IconCollection.get('compact_camera.svg'))
        self.action_camera_properties.setShortcut('Ctrl+C')

        self.action_plan_view = QAction('&Plan view', parent)
        self.action_plan_view.setIcon(IconCollection.get('upload.svg'))
        self.action_plan_view.setShortcut('Ctrl+1')

        self.action_north_view = QAction('&North view', parent)
        self.action_north_view.setIcon(IconCollection.get('up.svg'))
        self.action_north_view.setShortcut('Ctrl+2')

        self.action_east_view = QAction('&East view', parent)
        self.action_east_view.setIcon(IconCollection.get('right.svg'))
        self.action_east_view.setShortcut('Ctrl+3')

        self.action_fit_to_screen = QAction('&Fit to screen', parent)
        self.action_fit_to_screen.setIcon(IconCollection.get('collect.svg'))
        self.action_fit_to_screen.setShortcut('Ctrl+F')

        self.action_perspective_projection = QAction('P&erspective projection', parent)
        self.action_perspective_projection.setIcon(IconCollection.get('ruler.svg'))
        self.action_perspective_projection.setShortcut('Ctrl+4')

        self.action_orthographic_projection = QAction('O&rthographic projection', parent)
        self.action_orthographic_projection.setIcon(IconCollection.get('ruler.svg'))
        self.action_orthographic_projection.setShortcut('Ctrl+5')

        self.action_take_screenshot = QAction('&Take screenshot', parent)
        self.action_take_screenshot.setIcon(IconCollection.get('webcam.svg'))
        self.action_take_screenshot.setShortcut('Shift+S')



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from qtpy.QtWidgets import QToolBar
from .actioncollection import ActionCollection


class ToolBar(QToolBar):
    def __init__(self, parent=None):
        QToolBar.__init__(self, parent)
        self.setAcceptDrops(True)

        self.action_collection = ActionCollection(self)
        self.add_actions()

    def add_actions(self) -> None:
        self.addAction(self.action_collection.action_show_tree)
        self.addAction(self.action_collection.action_camera_properties)
        self.addSeparator()
        self.addAction(self.action_collection.action_plan_view)
        self.addAction(self.action_collection.action_north_view)
        self.addAction(self.action_collection.action_east_view)
        self.addAction(self.action_collection.action_fit_to_screen)
        self.addSeparator()
        self.addAction(self.action_collection.action_autofit)
        self.addAction(self.action_collection.action_turbo_rendering)
        self.addSeparator()
        self.addAction(self.action_collection.action_take_screenshot)

    def auto_connect(self, tree, viewer, camera) -> None:
        self.connect_tree(tree)
        self.connect_viewer(viewer)
        self.connect_camera(camera)
        tree.connect_viewer(viewer)
        camera.connect_viewer(viewer)

    def connect_tree(self, tree) -> None:
        actions = self.action_collection
        actions.action_show_tree.triggered.connect(tree.show)

    def connect_camera(self, camera):
        actions = self.action_collection
        actions.action_camera_properties.triggered.connect(camera.show)

    def connect_viewer(self, viewer) -> None:
        actions = self.action_collection
        actions.action_plan_view.triggered.connect(viewer.plan_view)
        actions.action_north_view.triggered.connect(viewer.north_view)
        actions.action_east_view.triggered.connect(viewer.east_view)
        actions.action_fit_to_screen.triggered.connect(viewer.fit_to_screen)

        actions.action_perspective_projection.triggered.connect(viewer.perspective_projection)
        actions.action_orthographic_projection.triggered.connect(viewer.orthographic_projection)

        # React to viewer changes ourselves, since blastsight>=0.5.1 will not auto-do it anymore
        def auto_fit() -> None:
            if actions.action_autofit.isChecked():
                viewer.fit_to_screen()

        def auto_turbo() -> None:
            viewer.set_turbo_rendering(actions.action_turbo_rendering.isChecked())

        actions.action_animate.triggered.connect(viewer.set_animated)
        actions.action_autofit.triggered.connect(auto_fit)
        actions.action_turbo_rendering.triggered.connect(auto_turbo)

        viewer.signal_file_modified.connect(auto_fit)
        viewer.signal_load_success.connect(auto_turbo)



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from collections import OrderedDict
from ..glprograms.shaderprogram import ShaderProgram
from ..drawables.gldrawable import GLDrawable


class GLCollection:
    def __init__(self, viewer=None):
        self._viewer = viewer
        self._programs = OrderedDict()
        self._collection = OrderedDict()
        self._uniform_data = {}
        self._needs_update = True

    """
    Drawable collection handlers
    """
    def add(self, drawable: GLDrawable) -> None:
        self._collection[drawable.id] = drawable
        self._needs_update = True

    def get(self, _id: int) -> GLDrawable:
        return self._collection.get(_id)

    def get_last(self) -> GLDrawable:
        return self.get(self.last_id)

    def get_all_ids(self) -> list:
        return list(self._collection.keys())

    def get_all_drawables(self) -> list:
        return list(self._collection.values())

    def delete(self, _id: int) -> None:
        drawable = self._collection.pop(_id)
        drawable.cleanup()

    def clear(self) -> None:
        self._collection.clear()

    def size(self) -> int:
        return len(self._collection)

    def filter(self, drawable_type: type) -> list:
        # The copy avoids RuntimeError: OrderedDict mutated during iteration
        return [x for x in self._collection.copy().values() if type(x) is drawable_type]

    def retrieve(self, drawable_type: type, required: str = 'all') -> callable:
        runner = {
            'all': lambda x: True,
            'mesh_standard': lambda x: not (x.is_turbo_ready or x.is_wireframed),
            'mesh_wireframe': lambda x: x.is_wireframed,
            'mesh_highlight': lambda x: x.is_highlighted,
            'mesh_turbo': lambda x: x.is_turbo_ready,
            'mesh_cross': lambda x: x.is_cross_sectionable,
            'block_legacy': lambda x: x.is_legacy,
            'block_standard': lambda x: not x.is_legacy,
        }

        return list(filter(runner.get(required), self.filter(drawable_type)))

    @property
    def last_id(self) -> int:
        # bool(dict) evaluates to False if the dictionary is empty
        return list(self._collection.keys())[-1] if bool(self._collection) else -1

    """
    ShaderProgram collection handlers
    """
    # FIXME We could have a collection handler, so every method from here and below
    #  could be in the GLCollectionHandler, instead of in a collection
    def initialize(self) -> None:
        for program in self.get_programs():
            program.initialize()

    def associate(self, program: ShaderProgram, retriever: callable) -> None:
        self._programs[program.get_base_name()] = {
            'program': program,
            'retriever': retriever,
        }

    def recreate(self) -> None:
        self._needs_update = True

    def get_programs(self) -> list:
        return [x.get('program') for x in self._programs.values()]

    def get_retrievers(self) -> list:
        return [x.get('retriever') for x in self._programs.values()]

    def update_drawables(self) -> None:
        # Update shader program so that it knows what to draw
        # FIXME Can we modify this method to not need self._needs_update ?
        if self._needs_update:
            for association in self._programs.values():
                program = association.get('program')
                retriever = association.get('retriever')
                program.set_drawables([d for d in retriever() if d.is_visible])
            self._needs_update = False

    def update_matrix(self, matrix: str, value) -> None:
        self._uniform_data[matrix] = value

    def update_uniform(self, program_name: str, uniform: str, *values) -> None:
        program = self._programs.get(program_name).get('program')
        program.update_uniform(uniform, *values)

    def update_uniforms(self) -> None:
        # Update matrices so that it knows where it's looking
        def bind_update(prog):
            prog.bind()
            for k, v in self._uniform_data.items():
                prog.update_uniform(k, v)

        # Apply to each program
        for program in self.get_programs():
            bind_update(program)

    """
    Drawing methods
    """
    def draw_opaques(self) -> None:
        for program in self.get_programs():
            if len(program.opaques) > 0:
                program.bind()
                program.draw()

    def draw_transparents(self) -> None:
        for program in self.get_programs():
            if len(program.transparents) > 0:
                program.bind()
                program.redraw()

    def draw(self) -> None:
        self.update_drawables()
        self.update_uniforms()
        self.draw_opaques()
        self.draw_transparents()



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import pathlib

from datetime import datetime
from qtpy.QtCore import QDir
from qtpy.QtCore import Qt
from qtpy.QtCore import QFileInfo
from qtpy.QtCore import QSettings
from qtpy.QtCore import QThreadPool
from qtpy.QtWidgets import QFileDialog
from qtpy.QtWidgets import QMainWindow
from qtpy.QtWidgets import QProgressBar

from .dialogs.helpdialog import HelpDialog
from .dialogs.aboutdialog import AboutDialog
from .iconcollection import IconCollection

from .tools import uic
from ..threadworker import ThreadWorker


class MainWindow(QMainWindow):
    def __init__(self, parent=None):
        QMainWindow.__init__(self, parent)
        uic.loadUi(f'{pathlib.Path(__file__).parent}/UI/mainwindow.ui', self)
        self.resize(1000, 620)

        self.setWindowTitle('BlastSight')
        self.setWindowIcon(IconCollection.get('blastsight.png'))
        self.toolbar.setWindowTitle('Toolbar')

        self.setFocusPolicy(Qt.StrongFocus)
        self.setAcceptDrops(True)
        self.statusBar.showMessage('Ready')

        # Attributes
        self.filters_dict = {
            'mesh': {
                'load': 'Mesh Files (*.dxf *.off *.h5m);;'
                        'DXF Files (*.dxf);;'
                        'OFF Files (*.off);;'
                        'H5M Files (*.h5m);;'
                        'All Files (*.*)',
                'export': 'BlastSight Mesh (*.h5m);;'
                          'OFF File (*.off);;'
            },
            'block': {
                'load': 'Block Files (*.csv *.h5p *.out);;'
                        'CSV Files (*.csv);;'
                        'H5P Files (*.h5p);;'
                        'GSLib Files (*.out);;'
                        'All Files (*.*)',
                'export': 'BlastSight Blocks (*.h5p);;'
                          'CSV File (*.csv);;'
            },
            'point': {
                'load': 'Point Files (*.csv *.h5p *.out);;'
                        'CSV Files (*.csv);;'
                        'H5P Files (*.h5p);;'
                        'GSLib Files (*.out);;'
                        'All Files (*.*)',
                'export': 'BlastSight Points (*.h5p);;'
                          'CSV File (*.csv);;'
            },
            'line': {
                'load': 'Line Files (*.csv *.dxf);;'
                        'CSV Files (*.csv);;'
                        'DXF Files (*.dxf);;'
                        'All Files (*.*)',
                'export': 'CSV File (*.csv);;'
            },
            'tube': {
                'load': 'Line Files (*.csv *.dxf);;'
                        'CSV Files (*.csv);;'
                        'DXF Files (*.dxf);;'
                        'All Files (*.*)',
                'export': 'CSV File (*.csv);;'
            }
        }

        self.settings = QSettings('BlastSight', application='blastsight', parent=self)

        # Progress bar (hidden by default)
        self.progress_bar = QProgressBar(self.statusBar)
        self.progress_bar.setValue(0)
        self.progress_bar.setMaximumWidth(self.width() / 5)
        self.progress_bar.hide()
        self.statusBar.addPermanentWidget(self.progress_bar)

        # Extra actions
        actions = self.toolbar.action_collection
        self.toolbar.addAction(actions.action_quit)
        self.setToolButtonStyle(Qt.ToolButtonTextUnderIcon)

        self.generate_menubar()
        self.connect_actions()

        # Set auto-fit as True when starting the app
        if not actions.action_autofit.isChecked():
            actions.action_autofit.trigger()

        # Set auto-rotate as True when starting the app
        if not actions.action_autorotate.isChecked():
            actions.action_autorotate.trigger()

        # Set animated as True when starting the app
        if not actions.action_animate.isChecked():
            actions.action_animate.trigger()

        # Set camera widget as hidden
        self.dockWidget_camera.hide()

        # self.title = self.windowTitle()
        # self.viewer.signal_fps_updated.connect(lambda x: self.setWindowTitle(f'{self.title} (FPS: {x:.1f})'))

    def generate_menubar(self) -> None:
        actions = self.toolbar.action_collection

        # File
        self.menu_File.addAction(actions.action_load_mesh)
        self.menu_File.addAction(actions.action_load_blocks)
        self.menu_File.addAction(actions.action_load_points)
        self.menu_File.addAction(actions.action_load_lines)
        self.menu_File.addAction(actions.action_load_tubes)
        self.menu_File.addSeparator()
        self.menu_File.addAction(actions.action_load_mesh_folder)
        self.menu_File.addAction(actions.action_load_blocks_folder)
        self.menu_File.addAction(actions.action_load_points_folder)
        self.menu_File.addAction(actions.action_load_lines_folder)
        self.menu_File.addAction(actions.action_load_tubes_folder)
        self.menu_File.addSeparator()
        self.menu_File.addAction(actions.action_quit)

        # View
        self.menu_View.addAction(actions.action_camera_properties)
        self.menu_View.addAction(actions.action_plan_view)
        self.menu_View.addAction(actions.action_north_view)
        self.menu_View.addAction(actions.action_east_view)
        self.menu_View.addAction(actions.action_fit_to_screen)
        self.menu_View.addSeparator()
        self.menu_View.addAction(actions.action_perspective_projection)
        self.menu_View.addAction(actions.action_orthographic_projection)
        self.menu_View.addSeparator()
        self.menu_View.addAction(actions.action_take_screenshot)

        # Tools
        self.menu_Tools.addAction(actions.action_slice_meshes)
        self.menu_Tools.addAction(actions.action_slice_blocks)
        self.menu_Tools.addAction(actions.action_detection_mode)
        self.menu_Tools.addAction(actions.action_measurement_mode)
        self.menu_Tools.addSeparator()
        self.menu_Tools.addAction(actions.action_normal_mode)
        self.menu_Tools.addAction(actions.action_cross_section)

        # Settings
        self.menu_Settings.addAction(actions.action_autofit)
        self.menu_Settings.addAction(actions.action_animate)
        self.menu_Settings.addAction(actions.action_autorotate)
        self.menu_Settings.addAction(actions.action_turbo_rendering)

        # Help
        self.menu_Help.addAction(actions.action_help)
        self.menu_Help.addAction(actions.action_about)

    def connect_actions(self) -> None:
        actions = self.toolbar.action_collection

        # Toolbar/Tree/Camera
        self.toolbar.connect_viewer(self.viewer)
        self.toolbar.connect_tree(self.dockWidget_tree)
        self.toolbar.connect_camera(self.dockWidget_camera)
        self.cameraWidget.connect_viewer(self.viewer)
        self.treeWidget.connect_viewer(self.viewer)
        self.treeWidget.enable_exportability(True)

        # File
        actions.action_load_mesh.triggered.connect(self.dialog_load_mesh)
        actions.action_load_blocks.triggered.connect(self.dialog_load_blocks)
        actions.action_load_points.triggered.connect(self.dialog_load_points)
        actions.action_load_lines.triggered.connect(self.dialog_load_lines)
        actions.action_load_tubes.triggered.connect(self.dialog_load_tubes)

        actions.action_load_mesh_folder.triggered.connect(self.dialog_load_mesh_folder)
        actions.action_load_blocks_folder.triggered.connect(self.dialog_load_blocks_folder)
        actions.action_load_points_folder.triggered.connect(self.dialog_load_points_folder)
        actions.action_load_lines_folder.triggered.connect(self.dialog_load_lines_folder)
        actions.action_load_tubes_folder.triggered.connect(self.dialog_load_tubes_folder)

        actions.action_quit.triggered.connect(self.close)

        # View
        actions.action_take_screenshot.triggered.connect(self.slot_screenshot)

        # Tools
        actions.action_slice_meshes.triggered.connect(self.slot_slice_meshes)
        actions.action_slice_blocks.triggered.connect(self.slot_slice_blocks)
        actions.action_detection_mode.triggered.connect(self.slot_detection_mode)
        actions.action_measurement_mode.triggered.connect(self.slot_measurement_mode)

        actions.action_normal_mode.triggered.connect(self.slot_normal_mode)
        actions.action_cross_section.triggered.connect(self.slot_cross_section)

        # Help
        actions.action_help.triggered.connect(self.slot_help)
        actions.action_about.triggered.connect(self.slot_about)

        # Viewer signals
        self.viewer.signal_fps_updated.connect(self.print_fps)
        self.viewer.signal_mode_updated.connect(self.slot_mode_updated)
        self.viewer.signal_elements_detected.connect(self.slot_elements_detected)
        self.viewer.signal_mesh_distances.connect(self.slot_mesh_distances)

        self.viewer.signal_load_success.connect(self.slot_element_load_success)
        self.viewer.signal_load_failure.connect(self.slot_element_load_failure)
        self.viewer.signal_export_success.connect(self.slot_element_export_success)
        self.viewer.signal_export_failure.connect(self.slot_element_export_failure)

        self.viewer.signal_process_updated.connect(self.slot_process_updated)
        self.viewer.signal_process_started.connect(self.slot_process_started)
        self.viewer.signal_process_finished.connect(self.slot_process_finished)

        # TreeWidget actions
        self.treeWidget.signal_export_element.connect(self._dialog_export_element)

    @property
    def last_dir(self) -> str:
        return self.settings.value('last_directory', '.')

    @last_dir.setter
    def last_dir(self, _last_dir: str) -> None:
        self.settings.setValue('last_directory', _last_dir)

    @staticmethod
    def print_fps(fps) -> None:
        print(f'               \r', end='')
        print(f'FPS: {fps:.1f} \r', end='')

    """
    Status bar updates
    """
    def slot_element_load_success(self, _id: int) -> None:
        self.statusBar.showMessage(f'Load successful (id: {_id}).')

    def slot_element_load_failure(self) -> None:
        self.statusBar.showMessage(f'Failed to load.')

    def slot_element_export_success(self, _id: int) -> None:
        self.statusBar.showMessage(f'Export successful (id: {_id}).')

    def slot_element_export_failure(self) -> None:
        self.statusBar.showMessage(f'Failed to export.')

    def slot_mode_updated(self, mode: str) -> None:
        self.statusBar.showMessage(f'Mode: {mode}')

    def slot_mesh_distances(self, distance_dict: dict) -> None:
        self.statusBar.showMessage(f'Distance: {distance_dict.get("distance")}')

    def slot_elements_detected(self, attributes: list) -> None:
        id_list = sorted(map(lambda attr: attr.get('id', -1), attributes))
        self.statusBar.showMessage(f'Detected elements: {id_list}')

    """
    Slots for cross-sections
    """
    def handle_slices(self, description: dict, executer: callable) -> None:
        # Retrieve description vectors
        origin = description.get('origin')
        normal = description.get('normal')
        up = description.get('up')

        # Execute the callable over the elements
        executer(origin, normal)

        # Auto-rotate camera to meet cross-section
        actions = self.toolbar.action_collection
        if actions.action_autorotate.isChecked():
            self.viewer.set_camera_from_vectors(normal, up)

    def slot_slice_meshes(self) -> None:
        def executer_standard(origin, normal) -> None:
            slices = self.viewer.slice_meshes(origin, normal)
            self.add_mesh_slices(slices)

        def executer_cross(origin, normal) -> None:
            self.viewer.cross_section(origin, normal)

        def handler(description: dict) -> None:
            if self.toolbar.action_collection.action_cross_section.isChecked():
                self.handle_slices(description, executer_cross)
            else:
                self.handle_slices(description, executer_standard)

            # Disconnect
            self.viewer.set_normal_mode()
            self.viewer.signal_slice_description.disconnect()

        self.viewer.set_slice_mode()
        self.viewer.signal_slice_description.connect(handler)

    def slot_slice_blocks(self) -> None:
        def executer(origin, normal) -> None:
            slices = self.viewer.slice_blocks(origin, normal)
            self.add_block_slices(slices)

        def handler(description: dict) -> None:
            self.handle_slices(description, executer)

            # Disconnect
            self.viewer.set_normal_mode()
            self.viewer.signal_slice_description.disconnect()

        self.viewer.set_slice_mode()
        self.viewer.signal_slice_description.connect(handler)

    def slot_cross_section(self, status: bool) -> None:
        self.viewer.set_cross_section(status)

        # Make all meshes either semi-transparent or fully opaque
        for drawable in self.viewer.get_all_drawables():
            drawable.alpha *= 0.1 if status else 10.0

        self.viewer.update_all()

    def add_mesh_slices(self, slice_list: list) -> None:
        def add_slice(description: dict) -> None:
            slices = description.get('vertices')
            mesh_id = description.get('element_id')
            mesh = self.viewer.get_drawable(mesh_id)

            def add_subslice(subslice, num: int) -> None:
                self.viewer.lines(vertices=subslice,
                                  color=mesh.color,
                                  name=f'MESHSLICE_{num}_{mesh.name}',
                                  extension='csv',
                                  loop=True)

            # Execute add_subslice for all subslices of the slice
            for i, ssl in enumerate(slices):
                add_subslice(ssl, i)

        # Execute add_slice over all slice descriptions
        for sl in slice_list:
            add_slice(sl)

    def add_block_slices(self, slice_list: list) -> None:
        def add_slice(description: dict) -> None:
            indices = description.get('indices')
            block_id = description.get('element_id')
            block = self.viewer.get_drawable(block_id)

            if 'SLICE' in str(block.name):
                return

            self.viewer.blocks(vertices=block.vertices[indices],
                               values=block.values[indices],
                               color=block.color[indices],
                               vmin=block.vmin,
                               vmax=block.vmax,
                               colormap=block.colormap,
                               block_size=block.block_size,
                               name=f'BLOCKSLICE_{block.name}',
                               extension='csv')

        # Execute add_slice over all slice descriptions
        for sl in slice_list:
            add_slice(sl)

    """
    Common functionality for loading
    """
    @staticmethod
    def _thread_runner(method: callable, *args, **kwargs) -> None:
        worker = ThreadWorker(method, *args, **kwargs)
        QThreadPool.globalInstance().start(worker)

    def _dialog_load_element(self, loader: classmethod, hint: str, *args, **kwargs) -> None:
        (paths, selected_filter) = QFileDialog.getOpenFileNames(
            parent=self,
            directory=self.last_dir,
            filter=self.filters_dict.get(hint).get('load'))

        path_list = sorted([p for p in paths if p != ''])
        if len(path_list) > 0:
            self.statusBar.showMessage(f'Loading {len(path_list)} element(s)...')
            self._thread_runner(self.viewer.load_multiple, path_list, loader, *args, **kwargs)
            self.last_dir = QFileInfo(path_list[-1]).absoluteDir().absolutePath()

    def _dialog_load_folder(self, loader: classmethod, *args, **kwargs) -> None:
        path = QFileDialog.getExistingDirectory(
            parent=self,
            directory=self.last_dir,
            options=QFileDialog.ShowDirsOnly)

        # Execute method
        if bool(path):
            self.statusBar.showMessage('Loading folder...')
            self._thread_runner(loader, path, *args, **kwargs)
            self.last_dir = path

    """
    Slots for progress updates
    """
    def slot_process_updated(self, value: int) -> None:
        self.progress_bar.setValue(value)

    def slot_process_started(self) -> None:
        self.progress_bar.setValue(0)
        self.progress_bar.show()

    def slot_process_finished(self,) -> None:
        self.progress_bar.hide()

    """
    Slot for screenshots
    """
    def slot_screenshot(self) -> None:
        (path, selected_filter) = QFileDialog.getSaveFileName(
            parent=self.viewer,
            directory=f'BlastSight Screenshot ({datetime.now().strftime("%Y%m%d-%H%M%S")})',
            filter='PNG image (*.png);;')

        self.viewer.take_screenshot(path)

    """
    Slots for loading files
    """
    def dialog_load_mesh(self) -> None:
        self._dialog_load_element(loader=self.viewer.load_mesh, hint='mesh')

    def dialog_load_blocks(self) -> None:
        self._dialog_load_element(loader=self.viewer.load_blocks, hint='block')

    def dialog_load_points(self) -> None:
        self._dialog_load_element(loader=self.viewer.load_points, hint='point')

    def dialog_load_lines(self) -> None:
        self._dialog_load_element(loader=self.viewer.load_lines, hint='line')

    def dialog_load_tubes(self) -> None:
        self._dialog_load_element(loader=self.viewer.load_tubes, hint='tube')

    def dialog_load_mesh_folder(self) -> None:
        self._dialog_load_folder(loader=self.viewer.load_mesh_folder)

    def dialog_load_blocks_folder(self) -> None:
        self._dialog_load_folder(loader=self.viewer.load_blocks_folder)

    def dialog_load_points_folder(self) -> None:
        self._dialog_load_folder(loader=self.viewer.load_points_folder)

    def dialog_load_lines_folder(self) -> None:
        self._dialog_load_folder(loader=self.viewer.load_lines_folder)

    def dialog_load_tubes_folder(self) -> None:
        self._dialog_load_folder(loader=self.viewer.load_tubes_folder)

    """
    Slots for exporting elements
    """
    def _dialog_export_element(self, _id: int) -> None:
        element = self.viewer.get_drawable(_id).element
        proposed_path = QDir(self.last_dir).filePath(element.name)

        # TODO Check hints by element type
        filters = 'BlastSight FIXME (*.h5m);;'

        (path, selected_filter) = QFileDialog.getSaveFileName(
            parent=self,
            directory=proposed_path,
            filter=filters)

        # Execute method
        if bool(path):
            self.statusBar.showMessage('Exporting...')
            self._thread_runner(self.viewer.export_element, path, _id)

    """
    Slots for modifying interaction modes
    """
    def slot_normal_mode(self) -> None:
        self.viewer.set_normal_mode()

    def slot_detection_mode(self) -> None:
        self.viewer.set_detection_mode()

    def slot_measurement_mode(self) -> None:
        self.viewer.set_measurement_mode()

    """
    Slots for showing help/about dialogs
    """
    def slot_help(self) -> None:
        HelpDialog(self).exec_()

    def slot_about(self) -> None:
        AboutDialog(self).exec_()

    """
    Events pass-through
    """
    def dragEnterEvent(self, event, *args, **kwargs) -> None:
        self.viewer.dragEnterEvent(event, *args, **kwargs)

    def dropEvent(self, event, *args, **kwargs) -> None:
        self.viewer.dropEvent(event, *args, **kwargs)



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import dxfgrabber
import numpy as np

from qtpy.QtCore import QFileInfo
from .parserdata import ParserData
from .parser import Parser


class DXFParser(Parser):
    @staticmethod
    def load_file(path: str, *args, **kwargs) -> ParserData:
        assert path.lower().endswith('dxf')

        dxf = dxfgrabber.readfile(path)

        hints = {'mesh': [dxfgrabber.dxfentities.Face],
                 'line': [dxfgrabber.dxfentities.Polyline, dxfgrabber.dxfentities.LWPolyline],
                 'tube': [dxfgrabber.dxfentities.Polyline, dxfgrabber.dxfentities.LWPolyline],
                 }
        hint = kwargs.get('hint', 'mesh')

        # Model data
        data = ParserData()

        # Detect vertices and indices
        entities = [e for e in dxf.entities if type(e) in hints.get(hint)]
        points = []

        for entity in entities:
            points.extend(entity.points[:3])

        if hint == 'mesh':
            vertices, indices = np.unique(np.array(points), axis=0, return_inverse=True)
            data.vertices = vertices
            data.indices = indices.reshape((-1, 3))
        else:
            data.vertices = np.array(points)

        # Metadata
        data.properties = {
            'name': QFileInfo(path).completeBaseName(),
            'extension': QFileInfo(path).suffix()
        }

        return data



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from .glcollection import GLCollection

from ..drawables.backgroundgl import BackgroundGL
from ..glprograms.backgroundprogram import BackgroundProgram


class GLBackgroundCollection(GLCollection):
    def __init__(self, viewer=None):
        super().__init__(viewer)
        self.associate(BackgroundProgram(viewer), lambda: self.retrieve(BackgroundGL))



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import numpy as np

from .gldrawable import GLDrawable
from OpenGL.GL import *


class PointGL(GLDrawable):
    def __init__(self, element, *args, **kwargs):
        super().__init__(element, *args, **kwargs)
        self.num_points = 0

    def setup_attributes(self) -> None:
        _POSITION = 0
        _COLOR = 1
        _ALPHA = 2
        _SIZE = 3

        # Generate VAO and VBOs (see GLDrawable)
        self.create_vao_vbos(4)

        # Data
        vertices = self.element.vertices.astype(np.float32)
        colors = self.element.color.astype(np.float32)
        alpha = np.array([self.element.alpha], np.float32)
        sizes = self.element.point_size.astype(np.float32)

        self.num_points = len(vertices)

        glBindVertexArray(self.vao)

        # Fill buffers (see GLDrawable)
        self.fill_buffer(_POSITION, 3, vertices, GLfloat, GL_FLOAT, self.vbos[_POSITION])
        self.fill_buffer(_COLOR, 3, colors, GLfloat, GL_FLOAT, self.vbos[_COLOR])
        self.fill_buffer(_ALPHA, 1, alpha, GLfloat, GL_FLOAT, self.vbos[_ALPHA])
        self.fill_buffer(_SIZE, 1, sizes, GLfloat, GL_FLOAT, self.vbos[_SIZE])

        # The attribute advances once per divisor instances of the set(s) of vertices being rendered
        # And guess what, we have just 1 instance, exactly what we wanted!
        glVertexAttribDivisor(_ALPHA, 1)

        glBindVertexArray(0)

    def draw(self) -> None:
        glBindVertexArray(self.vao)
        glDrawArrays(GL_POINTS, 0, self.num_points)
        glBindVertexArray(0)



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from .shaderprogram import ShaderProgram


class BlockProgram(ShaderProgram):
    def __init__(self, viewer):
        super().__init__(viewer)
        self.base_name = 'Block'

    def initialize(self) -> None:
        super().initialize()
        self.add_uniform_loc('block_size')

    def setup_shaders(self) -> None:
        # Placeholders to avoid early garbage collection
        vs = self.enable_vertex_shader()
        fs = self.enable_fragment_shader()
        gs = self.enable_geometry_shader()

        self.shader_program.link()

    def inner_draw(self, drawables: list) -> None:
        for drawable in drawables:
            self.update_uniform('block_size', *drawable.element.block_size)
            drawable.draw()



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from qtpy.QtCore import QRunnable


class ThreadWorker(QRunnable):
    def __init__(self, method, *args, **kwargs):
        super().__init__()
        self.method = method
        self.callback = kwargs.pop('callback', lambda *a, **k: None)
        self.args = args
        self.kwargs = kwargs

    def run(self) -> None:
        self.callback(self.method(*self.args, **self.kwargs))



#!/usr/bin/env python

"""
Adapted from https://github.com/mottosso/Qt.py

Dynamically loads an UI file using either PyQt5 or PySide2.
At least works better than qtpy for the PySide2 binding.
"""


def loadUi(uifile, baseinstance=None):
    try:
        # Return PyQt5.ui.loadUi directly
        from PyQt5 import uic

        return uic.loadUi(uifile, baseinstance)
    except (ImportError, TypeError):
        # Implement `PyQt5.uic.loadUi` for PySide(2)

        # ImportError: We'll use PySide2 because PyQt5 is not installed.
        # TypeError: We'll use PySide2 because the code uses it, but PyQt5 is also installed.

        import os
        import importlib

        from PySide2.QtUiTools import QUiLoader
        from PySide2 import QtCore
        from PySide2 import QtWidgets
        from xml.etree import ElementTree
        from io import StringIO

        class _UiLoader(QUiLoader):
            """Create the user interface in a base instance.

            Unlike `QUiLoader` itself this class does not
            create a new instance of the top-level widget, but creates the user
            interface in an existing instance of the top-level class if needed.

            This mimics the behaviour of `PyQt5.uic.loadUi`.

            """

            def __init__(self, baseinstance):
                super(_UiLoader, self).__init__(baseinstance)
                self.baseinstance = baseinstance
                self.custom_widgets = {}

            def _loadCustomWidgets(self, etree):
                """
                Workaround to pyside-77 bug.

                From QUiLoader doc we should use registerCustomWidget method.
                But this causes a segfault on some platforms.

                Instead we fetch from customwidgets DOM node the python class
                objects. Then we can directly use them in createWidget method.
                """

                def headerToModule(header):
                    """
                    Translate a header file to python module path
                    foo/bar.h => foo.bar
                    """
                    # Remove header extension
                    module = os.path.splitext(header)[0]

                    # Replace os separator by python module separator
                    return module.replace("/", ".").replace("\\", ".")

                custom_widgets = etree.find("customwidgets")

                if custom_widgets is None:
                    return

                for custom_widget in custom_widgets:
                    class_name = custom_widget.find("class").text
                    header = custom_widget.find("header").text
                    module = importlib.import_module(headerToModule(header))
                    self.custom_widgets[class_name] = getattr(module, class_name)

            def load(self, uifile, *args, **kwargs):
                from xml.etree.ElementTree import ElementTree

                # For whatever reason, if this doesn't happen then
                # reading an invalid or non-existing .ui file throws
                # a RuntimeError.
                etree = ElementTree()
                etree.parse(uifile)
                self._loadCustomWidgets(etree)

                widget = QUiLoader.load(self, uifile, *args, **kwargs)

                # Workaround for PySide 1.0.9, see issue #208
                widget.parentWidget()

                return widget

            def createWidget(self, class_name, parent=None, name=""):
                """Called for each widget defined in ui file

                Overridden here to populate `baseinstance` instead.

                """

                if parent is None and self.baseinstance:
                    # Supposed to create the top-level widget,
                    # return the base instance instead
                    return self.baseinstance

                # For some reason, Line is not in the list of available
                # widgets, but works fine, so we have to special case it here.
                if class_name in self.availableWidgets() + ["Line"]:
                    # Create a new widget for child widgets
                    widget = QUiLoader.createWidget(self, class_name, parent, name)
                elif class_name in self.custom_widgets:
                    widget = self.custom_widgets[class_name](parent)
                else:
                    raise Exception("Custom widget '%s' not supported"
                                    % class_name)

                if self.baseinstance:
                    # Set an attribute for the new child widget on the base
                    # instance, just like PyQt5.uic.loadUi does.
                    setattr(self.baseinstance, name, widget)

                return widget

        widget = _UiLoader(baseinstance).load(uifile)
        QtCore.QMetaObject.connectSlotsByName(widget)

        return widget



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from qtpy.QtCore import Qt, QPoint
from qtpy.QtGui import QMouseEvent
from .mode import Mode


class NormalMode(Mode):
    def __init__(self, viewer):
        super().__init__(viewer)
        self.name = 'Normal Mode'
        self.lastPos = None

    def mousePressEvent(self, event: QMouseEvent) -> None:
        super().mousePressEvent(event)
        self.lastPos = QPoint(event.pos())

    def mouseDoubleClickEvent(self, event: QMouseEvent) -> None:
        x, y, z = [event.pos().x(), event.pos().y(), 1]
        origin = self.viewer.origin_from_click(x, y, z)
        ray = self.viewer.ray_from_click(x, y, z)

        self.viewer.intersect_elements(origin, ray)

    def mouseMoveEvent(self, event: QMouseEvent) -> None:
        if self.lastPos is None:
            return

        dx = event.x() - self.lastPos.x()
        dy = event.y() - self.lastPos.y()
        smoothness = max(self.viewer.smoothness, 0.1)

        # viewer.world.setToIdentity()
        if event.buttons() == Qt.LeftButton:
            self.viewer.rotate(alpha=dy / smoothness, beta=0, gamma=dx / smoothness)

        elif event.buttons() == Qt.RightButton:
            self.viewer.rotate(alpha=dy / smoothness, beta=dx / smoothness, gamma=0)

        elif event.buttons() == Qt.MiddleButton:
            off_center = self.viewer.off_center[2]
            distance_x = off_center / self.viewer.width()
            distance_y = off_center / self.viewer.height()

            self.viewer.translate(x=-distance_x * dx, y=distance_y * dy, z=0)

        self.lastPos = QPoint(event.pos())



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from .elements.nullelement import NullElement
from .elements.meshelement import MeshElement
from .elements.blockelement import BlockElement
from .elements.pointelement import PointElement
from .elements.lineelement import LineElement
from .elements.tubeelement import TubeElement


class ElementFactory:
    @staticmethod
    def null(*args, **kwargs) -> NullElement:
        return NullElement(hint='null', *args, **kwargs)

    @staticmethod
    def mesh(*args, **kwargs) -> MeshElement:
        return MeshElement(hint='mesh', *args, **kwargs)

    @staticmethod
    def blocks(*args, **kwargs) -> BlockElement:
        return BlockElement(hint='block', *args, **kwargs)

    @staticmethod
    def points(*args, **kwargs) -> PointElement:
        return PointElement(hint='point', *args, **kwargs)

    @staticmethod
    def lines(*args, **kwargs) -> LineElement:
        return LineElement(hint='line', *args, **kwargs)

    @staticmethod
    def tubes(*args, **kwargs) -> TubeElement:
        return TubeElement(hint='tube', *args, **kwargs)



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from qtpy.QtGui import QFont
from qtpy.QtWidgets import QTreeWidgetItem


class TreeWidgetItem(QTreeWidgetItem):
    def __init__(self, parent=None, drawable=None):
        super().__init__(parent)
        self.drawable = drawable

        self.setText(0, f'{drawable.name}.{drawable.extension} (id: {drawable.id})')
        self.set_visible(drawable.is_visible)

    @property
    def id(self) -> int:
        return self.drawable.id

    def set_visible(self, is_visible: bool) -> None:
        font = QFont()
        font.setBold(is_visible)
        font.setItalic(not is_visible)
        self.setFont(0, font)

    def show(self) -> None:
        self.drawable.show()
        self.set_visible(self.drawable.is_visible)

    def hide(self) -> None:
        self.drawable.hide()
        self.set_visible(self.drawable.is_visible)

    def toggle_highlighting(self) -> None:
        self.drawable.toggle_highlighting()

    def toggle_wireframe(self) -> None:
        self.drawable.toggle_wireframe()

    def toggle_visibility(self) -> None:
        self.drawable.toggle_visibility()
        self.set_visible(self.drawable.is_visible)



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import numpy as np
import warnings

from . import utils

from qtpy.QtCore import QDirIterator
from qtpy.QtCore import QFileInfo
from qtpy.QtCore import QMutex
from qtpy.QtCore import QMutexLocker

from .elementfactory import ElementFactory

from .elements.element import Element
from .elements.elementcollection import ElementCollection
from .elements.blockelement import BlockElement
from .elements.pointelement import PointElement
from .elements.lineelement import LineElement
from .elements.meshelement import MeshElement
from .elements.nullelement import NullElement
from .elements.tubeelement import TubeElement

from .parsers.parser import Parser
from .parsers.parsercollection import ParserCollection
from .parsers.dxfparser import DXFParser
from .parsers.offparser import OFFParser
from .parsers.h5mparser import H5MParser
from .parsers.h5pparser import H5PParser
from .parsers.csvparser import CSVParser
from .parsers.gslibparser import GSLibParser


class Model:
    def __init__(self):
        self._mutex = QMutex()
        self.parser_collection = ParserCollection()
        self.element_collection = ElementCollection()
        self.factory = ElementFactory()

        self.add_parser('dxf', DXFParser())
        self.add_parser('off', OFFParser())
        self.add_parser('h5m', H5MParser())
        self.add_parser('h5p', H5PParser())
        self.add_parser('csv', CSVParser())
        self.add_parser('out', GSLibParser())

    @property
    def last_id(self) -> int:
        return self.element_collection.last_id

    """
    Utilities
    """
    def add_parser(self, extension: str, handler: Parser) -> None:
        self.parser_collection.add(extension, handler)

    def get_parser(self, extension: str) -> Parser:
        return self.parser_collection.get(extension)

    @staticmethod
    def get_paths_from_directory(path: str) -> list:
        it = QDirIterator(path, QDirIterator.Subdirectories)
        path_list = []

        while it.hasNext():
            next_path = it.next()
            if QFileInfo(next_path).isFile():
                path_list.append(next_path)

        return sorted(path_list)

    """
    Register methods
    """
    def register_element(self, element):
        # In a multi-threaded application, we can't risk assigning
        # the same ID to different elements, so we use a mutex here.
        with QMutexLocker(self._mutex):
            self.element_collection.add(element)

        return element

    def register_element_by_path(self, path: str, generator, *args, **kwargs):
        ext = path.split('.')[-1]
        info = self.get_parser(ext).load_file(path, *args, **kwargs)
        data = info.data
        properties = info.properties

        # Prioritize kwargs over file properties (useful in CLI)
        kwargs['data'] = kwargs.get('data', data)
        for k, v in properties.items():
            kwargs[k] = kwargs.get(k, v)

        return self.register_element(generator(*args, **kwargs))

    """
    Load methods by arguments
    """
    def null(self, *args, **kwargs) -> NullElement:
        # A NullElement won't be registered
        return self.factory.null(*args, **kwargs)

    def mesh(self, *args, **kwargs) -> MeshElement:
        return self.register_element(self.factory.mesh(*args, **kwargs))

    def blocks(self, *args, **kwargs) -> BlockElement:
        return self.register_element(self.factory.blocks(*args, **kwargs))

    def points(self, *args, **kwargs) -> PointElement:
        return self.register_element(self.factory.points(*args, **kwargs))

    def lines(self, *args, **kwargs) -> LineElement:
        return self.register_element(self.factory.lines(*args, **kwargs))

    def tubes(self, *args, **kwargs) -> TubeElement:
        return self.register_element(self.factory.tubes(*args, **kwargs))

    """
    Load methods by path
    """
    def load_mesh(self, path: str, *args, **kwargs) -> MeshElement:
        return self.register_element_by_path(path, self.factory.mesh, *args, **kwargs)

    def load_blocks(self, path: str, *args, **kwargs) -> BlockElement:
        return self.register_element_by_path(path, self.factory.blocks, *args, **kwargs)

    def load_points(self, path: str, *args, **kwargs) -> PointElement:
        return self.register_element_by_path(path, self.factory.points, *args, **kwargs)

    def load_lines(self, path: str, *args, **kwargs) -> LineElement:
        return self.register_element_by_path(path, self.factory.lines, *args, **kwargs)

    def load_tubes(self, path: str, *args, **kwargs) -> TubeElement:
        return self.register_element_by_path(path, self.factory.tubes, *args, **kwargs)

    """
    Element handling
    """
    def get(self, _id: int) -> Element:
        return self.element_collection.get(_id)

    def delete(self, _id: int) -> None:
        self.element_collection.delete(_id)

    """
    Element exporting
    """
    def export(self, path: str, _id: int) -> None:
        element = self.get(_id)
        ext = path.split('.')[-1]

        data = element.data
        properties = {}
        for k in element.exportable_properties:
            properties[k] = getattr(element, k)

        self.get_parser(ext).save_file(path=path, data=data, properties=properties)

    """
    Adapter for viewer's utilities (slice/distance)
    """
    @staticmethod
    def slice_meshes(origin: np.ndarray, normal: np.ndarray, meshes: list) -> list:
        """
        Returns a list of dicts, where each dict is the mesh ID and its sliced vertices
        """
        result = []

        for mesh in meshes:
            vertices = mesh.slice_with_plane(origin, normal)
            if len(vertices) > 0:
                result.append({'element_id': mesh.id,
                               'vertices': vertices,
                               })

        return result

    @staticmethod
    def slice_blocks(origin: np.ndarray, normal: np.ndarray, block_list: list) -> list:
        """
        Returns a list of dicts, where each dict is the block ID and its sliced indices
        """
        result = []

        for block in block_list:
            indices = block.slice_with_plane(origin, normal)
            if len(indices) > 0:
                result.append({
                    'element_id': block.id,
                    'indices': indices,
                })

        return result

    @staticmethod
    def measure_from_rays(origin_list: list, ray_list: list, meshes: list) -> dict:
        """
        Returns a dict with the following structure:

        {
            'point_a': list(float) or None,
            'point_b': list(float) or None',
            'distance': float or None
        }
        """
        points_A = []
        points_B = []

        closest_A = None
        closest_B = None

        # Detect intersections
        for mesh in meshes:
            int_A = mesh.intersect_with_ray(origin_list[0], ray_list[0])
            int_B = mesh.intersect_with_ray(origin_list[1], ray_list[1])

            # Discard non-intersections
            if int_A.size > 0:
                points_A.append(utils.closest_point_to(origin_list[0], int_A))

            if int_B.size > 0:
                points_B.append(utils.closest_point_to(origin_list[1], int_B))

        # Get closest points for each origin
        if len(points_A) > 0:
            points_A = np.vstack(points_A)
            closest_A = utils.closest_point_to(origin_list[0], points_A)

        if len(points_B) > 0:
            points_B = np.vstack(points_B)
            closest_B = utils.closest_point_to(origin_list[1], points_B)

        # Calculate distance if possible
        try:
            distance = utils.magnitude(closest_B - closest_A)
        except TypeError:
            distance = None

        return {
            'point_a': closest_A,
            'point_b': closest_B,
            'distance': distance,
        }

    @staticmethod
    def intersect_meshes(origin: np.ndarray, ray: np.ndarray, meshes: list) -> list:
        attributes_list = []

        for mesh in meshes:
            intersections = mesh.intersect_with_ray(origin, ray)
            closest_point = utils.closest_point_to(origin, intersections)
            if closest_point is not None:
                attributes = {**mesh.attributes,
                              'intersections': intersections,
                              'closest_point': closest_point,
                              }
                attributes_list.append(attributes)

        def sort_by_distance(x: dict) -> float:
            return utils.magnitude(x.get('closest_point') - origin)

        return sorted(attributes_list, key=sort_by_distance)

    @staticmethod
    def intersect_lines(origin: np.ndarray, ray: np.ndarray, lines: list) -> list:
        attributes_list = []

        for line in lines:
            intersections = line.intersect_with_ray(origin, ray)
            closest_point = utils.closest_point_to(origin, intersections)
            if closest_point is not None:
                attributes = {**line.attributes,
                              'intersections': intersections,
                              'closest_point': closest_point,
                              }
                attributes_list.append(attributes)

        return attributes_list



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import numpy as np

from .gldrawable import GLDrawable
from OpenGL.GL import *


class TubeGL(GLDrawable):
    def __init__(self, element, *args, **kwargs):
        super().__init__(element, *args, **kwargs)

        self.indices_size = 0

    def setup_attributes(self) -> None:
        _POSITION = 0
        _COLOR = 1

        # Generate VAO and VBOs (see GLDrawable)
        self.create_vao_vbos(3)

        # Data
        vertices, indices = self.element.as_mesh()

        vertices = np.array(vertices).astype(np.float32)
        indices = np.array(indices).astype(np.uint32)
        colors = self.element.rgba.astype(np.float32)

        self.indices_size = indices.size

        glBindVertexArray(self.vao)

        # Fill buffers (see GLDrawable)
        self.fill_buffer(_POSITION, 3, vertices, GLfloat, GL_FLOAT, self.vbos[_POSITION])
        self.fill_buffer(_COLOR, 4, colors, GLfloat, GL_FLOAT, self.vbos[_COLOR])

        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self.vbos[-1])
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, indices, GL_STATIC_DRAW)

        # The attribute advances once per divisor instances of the set(s) of vertices being rendered
        # And guess what, we have just 1 instance, exactly what we wanted!
        glVertexAttribDivisor(_COLOR, 1)

        glBindVertexArray(0)

    def draw(self) -> None:
        glBindVertexArray(self.vao)
        glDrawElements(GL_TRIANGLES, self.indices_size, GL_UNSIGNED_INT, None)
        glBindVertexArray(0)



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import pandas as pd
from qtpy.QtCore import QFileInfo
from .parserdata import ParserData
from .parser import Parser


class GSLibParser(Parser):
    @staticmethod
    def load_file(path: str, *args, **kwargs) -> ParserData:
        assert path.lower().endswith('out')

        # Metadata
        properties = {
            'name': QFileInfo(path).completeBaseName(),
            'extension': QFileInfo(path).suffix()
        }

        data = ParserData()
        try:
            header_count, headers = GSLibParser.get_header_info(path)
            with open(path, 'r') as f:
                data.data = pd.read_csv(f, sep=' ', header=None, names=headers, skiprows=header_count + 2)
                data.properties = properties
        except Exception:  # This "GSLib" file doesn't have headers
            print(f'*** WARNING: We can read {path}, but keep in mind that this not a real GSLib file. ***')
            with open(path, 'r') as f:
                data.data = pd.read_csv(f, sep=' ', header=None, prefix='col_')
                data.properties = properties

        return data

    @staticmethod
    def get_header_info(path: str) -> tuple:
        with open(path, 'r') as gslib_file:
            gslib_file.readline()
            header_count = int(gslib_file.readline().strip().split()[0])
            headers = [gslib_file.readline().strip().split()[0] for _ in range(header_count)]
            return header_count, headers



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

# Solving memory leak problem in pandas
# https://github.com/pandas-dev/pandas/issues/2659#issuecomment-415177442

import pandas as pd
from ctypes import cdll, CDLL

try:
    cdll.LoadLibrary('libc.so.6')
    libc = CDLL('libc.so.6')
    libc.malloc_trim(0)
except (OSError, AttributeError):
    libc = None

__old_del = getattr(pd.DataFrame, '__del__', None)


def __new_del(self):
    if __old_del:
        __old_del(self)
    libc.malloc_trim(0)


if libc:
    pd.DataFrame.__del__ = __new_del



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from .shaderprogram import ShaderProgram
from OpenGL.GL import *


class HighlightProgram(ShaderProgram):
    def __init__(self, viewer):
        super().__init__(viewer)
        self.base_name = 'Highlight'

    def inner_draw(self, drawables: list) -> None:
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)
        glLineWidth(3)
        super().inner_draw(drawables)
        glLineWidth(1)
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL)



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import json

from qtpy.QtCore import Qt
from qtpy.QtCore import Signal
from qtpy.QtGui import QKeyEvent
from qtpy.QtWidgets import QAbstractItemView
from qtpy.QtWidgets import QMenu
from qtpy.QtWidgets import QTreeWidget
from qtpy.QtWidgets import QTreeWidgetItemIterator

from .treewidgetitem import TreeWidgetItem
from .actioncollection import ActionCollection

from .dialogs.colordialog import ColorDialog
from .dialogs.propertiesdialog import PropertiesDialog

from ..drawables.meshgl import MeshGL
from ..drawables.blockgl import BlockGL
from ..drawables.pointgl import PointGL
from ..drawables.linegl import LineGL
from ..drawables.tubegl import TubeGL


class TreeWidget(QTreeWidget):
    signal_export_element = Signal(int)

    def __init__(self, viewer=None, parent=None):
        super().__init__(parent)
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.setWindowFlags(Qt.WindowStaysOnTopHint)
        self.setSelectionMode(QAbstractItemView.ExtendedSelection)

        self.itemDoubleClicked.connect(self.center_camera)
        self.customContextMenuRequested.connect(self.context_menu)

        self.setWindowTitle('Element list')
        self.headerItem().setText(0, 'Elements')

        self.is_export_enabled = False
        self.viewer = viewer

    def connect_viewer(self, viewer=None) -> None:
        if bool(viewer):
            self.viewer = viewer
        if not bool(self.viewer):
            return

        self.viewer.signal_file_modified.connect(self.auto_refill)
        self.viewer.signal_elements_detected.connect(self.handle_elements_detected)

    def handle_elements_detected(self, attributes: list) -> None:
        self.select_by_id_list(list(map(lambda attr: attr.get('id', -1), attributes)))

    def handle_color(self, item: TreeWidgetItem) -> None:
        element = self.viewer.get_drawable(item.id)
        dialog = ColorDialog(element)

        def update_color() -> None:
            element.rgba = dialog.currentColor().getRgbF()
            self.viewer.update_drawable(element.id)

        dialog.accepted.connect(update_color)
        dialog.show()

    def handle_multiple(self, item_list: list) -> None:
        dialog = ColorDialog()

        def update_color(item: TreeWidgetItem) -> None:
            element = self.viewer.get_drawable(item.id)
            element.rgba = dialog.currentColor().getRgbF()
            self.viewer.update_drawable(element.id)

        def update_multiple():
            for item in item_list:
                update_color(item)

        dialog.accepted.connect(update_multiple)
        dialog.show()

    def handle_properties(self, item: TreeWidgetItem) -> None:
        element = self.viewer.get_drawable(item.id)
        dialog = PropertiesDialog(element)
        dialog.accepted.connect(lambda: self.update_properties(dialog, element))
        dialog.show()

    def update_properties(self, dialog, element) -> None:
        # Parse values in QTableWidget
        for i in range(dialog.tableWidget_properties.rowCount()):
            k = dialog.tableWidget_properties.verticalHeaderItem(i).text()
            v = dialog.tableWidget_properties.item(i, 0).text()
            try:
                setattr(element, k, float(v))
            except ValueError:  # Element might be a list or string
                try:  # Element is a list
                    setattr(element, k, json.loads(v))
                except json.decoder.JSONDecodeError:  # Element is a string
                    setattr(element, k, v)
            except KeyError:  # Element clearly is not a property
                print(f'{k} property does not exist.')

        # Update headers
        altered = dialog.has_altered_coordinates(element)
        element.headers = dialog.current_headers

        # Recreate instance with the "new" data
        self.viewer.update_drawable(element.id)

        # If coordinates were altered and auto-fit is enabled, call fit_to_screen()
        if altered and self.viewer.get_autofit_status():
            self.viewer.fit_to_screen()

    def enable_exportability(self, value: bool) -> None:
        self.is_export_enabled = value

    def auto_refill(self) -> None:
        self.clear()

        for drawable in self.available_drawables():
            self.addTopLevelItem(self.generate_item(drawable))

    def available_drawables(self) -> list:
        return self.viewer.get_all_drawables()

    def generate_item(self, drawable) -> TreeWidgetItem:
        return TreeWidgetItem(self, drawable)

    def context_menu(self, event) -> None:
        # Pop-up the context menu on current position, if an item is there
        item = self.itemAt(event)

        if item:
            self.show_context_menu(item, self.viewport().mapToGlobal(event))

    def connect_actions(self, actions: ActionCollection, item: TreeWidgetItem) -> None:
        self._connect_standalone_actions(actions)
        self._connect_item_actions(actions, item)

    def _connect_standalone_actions(self, actions: ActionCollection) -> None:
        actions.action_show.triggered.connect(self.show_items)
        actions.action_hide.triggered.connect(self.hide_items)
        actions.action_delete.triggered.connect(self.delete_items)
        actions.action_focus_camera.triggered.connect(self.center_camera)

    def _connect_item_actions(self, actions: ActionCollection, item: TreeWidgetItem) -> None:
        actions.action_highlight.triggered.connect(item.toggle_highlighting)
        actions.action_wireframe.triggered.connect(item.toggle_wireframe)

        actions.action_setup_colors.triggered.connect(lambda: self.handle_color(item))
        actions.action_properties.triggered.connect(lambda: self.handle_properties(item))

        actions.action_export_element.triggered.connect(lambda: self.signal_export_element.emit(item.id))

    def generate_multiple_menu(self, action_squeezer: callable = lambda *args: None) -> QMenu:
        menu = QMenu()
        actions = ActionCollection(self)

        self._connect_standalone_actions(actions)
        actions.action_setup_multiple.triggered.connect(
            lambda: self.handle_multiple(self.selectedItems()))

        menu.addAction(actions.action_show)
        menu.addAction(actions.action_hide)
        menu.addSeparator()

        # Call the method that squeezes more actions before the export/delete ones
        action_squeezer(menu, actions)

        menu.addAction(actions.action_delete)

        return menu

    def generate_single_menu(self, item: TreeWidgetItem, action_squeezer: callable) -> QMenu:
        menu = QMenu()
        actions = ActionCollection(self)
        self.connect_actions(actions, item)

        menu.addAction(actions.action_show)
        menu.addAction(actions.action_hide)
        menu.addAction(actions.action_focus_camera)
        menu.addSeparator()

        # Call the method that squeezes more actions before the export/delete ones
        action_squeezer(menu, actions)

        if self.is_export_enabled:
            menu.addAction(actions.action_export_element)
        menu.addAction(actions.action_delete)

        return menu

    def generate_mesh_menu(self, item: TreeWidgetItem) -> QMenu:
        def action_squeezer(menu: QMenu, actions: ActionCollection) -> None:
            # Dynamic checkbox in actions
            actions.action_highlight.setChecked(item.drawable.is_highlighted)
            actions.action_wireframe.setChecked(item.drawable.is_wireframed)

            menu.addAction(actions.action_highlight)
            menu.addAction(actions.action_wireframe)

            menu.addAction(actions.action_setup_colors)
            menu.addSeparator()

        return self.generate_single_menu(item, action_squeezer)

    def generate_standard_menu(self, item: TreeWidgetItem) -> QMenu:
        def action_squeezer(menu: QMenu, actions: ActionCollection) -> None:
            menu.addAction(actions.action_setup_colors)
            menu.addSeparator()

        return self.generate_single_menu(item, action_squeezer)

    def generate_dataframe_menu(self, item: TreeWidgetItem) -> QMenu:
        def action_squeezer(menu: QMenu, actions: ActionCollection) -> None:
            menu.addAction(actions.action_properties)
            menu.addSeparator()

        return self.generate_single_menu(item, action_squeezer)

    def show_context_menu(self, item, global_pos) -> None:
        menus_dict = {
            MeshGL: lambda: self.generate_mesh_menu(item),
            BlockGL: lambda: self.generate_dataframe_menu(item),
            PointGL: lambda: self.generate_dataframe_menu(item),
            LineGL: lambda: self.generate_standard_menu(item),
            TubeGL: lambda: self.generate_standard_menu(item),
        }

        def action_squeezer(menu: QMenu, actions: ActionCollection) -> None:
            menu.addAction(actions.action_setup_multiple)
            menu.addSeparator()

        # If multiple elements are selected, we'll only show a basic menu
        if len(self.selectedItems()) > 1:
            if all(map(lambda it: type(it.drawable) in [MeshGL, LineGL, TubeGL], self.selectedItems())):
                ctx_menu = self.generate_multiple_menu(action_squeezer)
            else:
                ctx_menu = self.generate_multiple_menu()
        else:
            ctx_menu = menus_dict.get(type(item.drawable))()

        ctx_menu.exec_(global_pos)

    def center_camera(self) -> None:
        row = min([self.indexOfTopLevelItem(x) for x in self.selectedItems()], default=0)
        drawable = self.topLevelItem(row).drawable
        self.viewer.show_drawable(drawable.id)
        self.viewer.camera_at(drawable.id)

    def select_item(self, row: int) -> None:
        row = max(min(row, self.topLevelItemCount() - 1), 0)
        item = self.topLevelItem(row)
        if item:
            item.setSelected(True)
            self.scrollToItem(item, QAbstractItemView.EnsureVisible)

    def select_by_id_list(self, id_list: list) -> None:
        self.clearSelection()
        it = QTreeWidgetItemIterator(self)

        first_item = False
        while it.value():
            item = it.value()
            if item.id in id_list:
                if not first_item:
                    self.setCurrentItem(item)
                    first_item = True

                self.select_item(self.indexOfTopLevelItem(item))
            it += 1

    def show_items(self) -> None:
        for item in self.selectedItems():
            item.show()

    def hide_items(self) -> None:
        for item in self.selectedItems():
            item.hide()

    def toggle_items_visibility(self) -> None:
        for item in self.selectedItems():
            item.toggle_visibility()

    def delete_items(self) -> None:
        closest_row = min([self.indexOfTopLevelItem(x) for x in self.selectedItems()], default=0)

        # We'll trick the viewer so it doesn't emit file_modified_signal until last item
        # has been deleted, so we're not forced to auto-fill for each deleted item.
        self.viewer.blockSignals(True)

        for _id in [item.drawable.id for item in self.selectedItems()]:
            self.viewer.delete(_id)

        self.viewer.blockSignals(False)
        self.viewer.signal_file_modified.emit()

        self.setCurrentItem(self.topLevelItem(max(closest_row - 1, 0)))

    def keyPressEvent(self, event: QKeyEvent) -> None:
        super().keyPressEvent(event)
        if self.topLevelItemCount() == 0:
            return

        shortcut_commands_dict = {
            Qt.Key_Space: self.toggle_items_visibility,
            Qt.Key_Delete: self.delete_items,
            Qt.Key_Enter: self.center_camera,
            Qt.Key_Return: self.center_camera,
        }

        # Execute command based on event.key()
        shortcut_commands_dict.get(event.key(), lambda: None)()



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from .parserdata import ParserData


class Parser:
    @staticmethod
    def load_file(path: str, *args, **kwargs) -> ParserData:
        raise NotImplementedError

    @staticmethod
    def save_file(path: str, *args, **kwargs) -> None:
        raise NotImplementedError



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import numpy as np
from .element import Element
from . import tubeutils


class TubeElement(Element):
    def __init__(self, *args, **kwargs):
        """
        TubeElement is a class inheriting from Element.

        {
            'data': {
                'x': list[float],
                'y': list[float],
                'z': list[float],
            }
            'properties': {
                'color': list[float],
                'alpha': float,
                'radius': float,
                'resolution': int
            },
            'metadata': {
                'id': int,
                'name': str or None,
                'extension': str or None
            }
        }
        """
        super().__init__(*args, **kwargs)

    def _fill_properties(self, *args, **kwargs) -> None:
        super()._fill_properties(*args, **kwargs)
        self.radius = kwargs.get('radius', 0.15)
        self.resolution = kwargs.get('resolution', 15)
        self.loop = kwargs.get('loop', False)

    def _check_integrity(self) -> None:
        super()._check_integrity()
        if len(self.vertices) < 2:
            raise ValueError("Not enough data to create this element.")

        # Append first vertex to self.vertices if a loop was enabled
        if self.loop:
            self.vertices = np.append(self.vertices, [self.vertices[0, :]], axis=0)

    """
    Properties
    """
    @property
    def radius(self) -> float:
        return self.properties.get('radius')

    @property
    def resolution(self) -> int:
        return self.properties.get('resolution')

    @property
    def loop(self) -> bool:
        return self.properties.get('loop')

    @radius.setter
    def radius(self, _radius: float) -> None:
        self.properties['radius'] = _radius

    @resolution.setter
    def resolution(self, _resolution: int) -> None:
        self.properties['resolution'] = _resolution

    @loop.setter
    def loop(self, _loop: bool) -> None:
        self.properties['loop'] = _loop

    """
    Utilities
    """
    def as_mesh(self) -> tuple:
        vertices = []
        indices = []
        delta = 0

        for v0, v1 in zip(self.vertices[:-1], self.vertices[1:]):
            c_vertices, c_faces = tubeutils.cylinder(self.radius, self.resolution, [v0, v1])
            vertices.append(c_vertices)
            indices.append(c_faces + delta)
            delta += (c_faces.max() + 1)

        return np.concatenate(vertices), np.concatenate(indices)



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from collections import OrderedDict
from .element import Element


class ElementCollection:
    def __init__(self):
        self._collection = OrderedDict()
        self._current_id = -1

    @property
    def last_id(self) -> int:
        # bool(dict) evaluates to False if the dictionary is empty
        return list(self._collection.keys())[-1] if bool(self._collection) else -1

    def add(self, element: Element) -> None:
        self._current_id += 1
        self._collection[self._current_id] = element
        element.id = self._current_id  # Auto-update on element add

    def get(self, _id: int) -> Element:
        return self._collection.get(_id)

    def size(self) -> int:
        return len(self._collection)

    def delete(self, _id: int) -> None:
        del self._collection[_id]



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from .element import Element
from .elementcollection import ElementCollection
from .blockelement import BlockElement
from .pointelement import PointElement
from .lineelement import LineElement
from .meshelement import MeshElement
from .tubeelement import TubeElement



#!/usr/bin/env python

# These methods are taken and adapted from PyMesh.
# https://github.com/PyMesh/PyMesh
#
# PyMesh's license is Mozilla Public License 2, and as such,
# this particular file must be licensed as MPL2.
#
# See https://www.mozilla.org/en-US/MPL/ for more info.

import numpy as np


def cylinder(radius: float, resolution: int, segment: tuple) -> tuple:
    Z = np.array([0.0, 0.0, 1.0])
    p0, p1 = segment

    axis = p1 - p0
    length = np.linalg.norm(axis)

    if length <= 1e-12:
        axis = Z

    angles = np.linspace(0, 2 * np.pi, resolution + 1)[:-1]
    rim = np.column_stack((np.sin(angles), np.cos(angles), np.zeros(angles.size)))
    rot = to_matrix(quat_from_data(Z, axis))

    bottom_rim = np.dot(rot, rim.T).T * radius + p0
    top_rim = np.dot(rot, rim.T).T * radius + p1

    vertices = np.vstack([[p0, p1], bottom_rim, top_rim])

    vec_resolution = np.arange(resolution)

    # Bottom fan
    bottom_fan = np.zeros((resolution, 3))
    bottom_fan[:, 0] = np.zeros(resolution)
    bottom_fan[:, 1] = np.remainder(vec_resolution + 1, resolution) + 2
    bottom_fan[:, 2] = vec_resolution + 2

    # Top fan
    top_fan = np.zeros((resolution, 3))
    top_fan[:, 0] = np.ones(resolution)
    top_fan[:, 1] = vec_resolution + resolution + 2
    top_fan[:, 2] = np.remainder(vec_resolution + 1, resolution) + resolution + 2

    # Sides
    side_a = np.zeros((resolution, 3))
    side_a[:, 0] = vec_resolution + 2
    side_a[:, 1] = np.remainder(vec_resolution + 1, resolution) + 2
    side_a[:, 2] = vec_resolution + 2 + resolution

    side_b = np.zeros((resolution, 3))
    side_b[:, 0] = vec_resolution + 2 + resolution
    side_b[:, 1] = np.remainder(vec_resolution + 1, resolution) + 2
    side_b[:, 2] = np.remainder(vec_resolution + 1, resolution) + 2 + resolution

    side = np.append(side_a, side_b).astype(np.int)
    side = side.reshape((-1, 3), order='C')

    indices = np.vstack([bottom_fan, top_fan, side])

    return vertices, indices


def quat_from_data(v1: np.ndarray, v2: np.ndarray) -> np.ndarray:
    eps = 1e-12
    v1 /= np.linalg.norm(v1)
    v2 /= np.linalg.norm(v2)
    c = np.dot(v1, v2)
    if c < -1.0 + eps:
        # v1 is parallel and opposite of v2
        u, s, v = np.linalg.svd(np.array([v1, v2]))
        axis = v[2, :]
    else:
        axis = np.cross(v1, v2)
        l = np.linalg.norm(axis)
        if l > 0.0:
            axis /= np.linalg.norm(axis)
        else:
            # Parallel vectors.
            axis = v1

    w_sq = 0.5 * (1.0 + c)
    l = np.sqrt(1.0 - w_sq) * axis
    quat = np.array([np.sqrt(w_sq), l[0], l[1], l[2]])

    return quat


def to_matrix(a: np.ndarray) -> np.ndarray:
    return np.array([
        [1 - 2 * a[2] * a[2] - 2 * a[3] * a[3], 2 * a[1] * a[2] - 2 * a[3] * a[0],
         2 * a[1] * a[3] + 2 * a[2] * a[0]],
        [2 * a[1] * a[2] + 2 * a[3] * a[0], 1 - 2 * a[1] * a[1] - 2 * a[3] * a[3],
         2 * a[2] * a[3] - 2 * a[1] * a[0]],
        [2 * a[1] * a[3] - 2 * a[2] * a[0], 2 * a[2] * a[3] + 2 * a[1] * a[0],
         1 - 2 * a[1] * a[1] - 2 * a[2] * a[2]],
    ])



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from qtpy.QtCore import Qt
from qtpy.QtGui import QMouseEvent
from .mode import Mode


class SliceMode(Mode):
    def __init__(self, viewer):
        super().__init__(viewer)
        self.name = 'Slice Mode'
        self.origins = []
        self.rays = []

    def mousePressEvent(self, event: QMouseEvent) -> None:
        super().mousePressEvent(event)

        if event.buttons() == Qt.LeftButton:
            self.detect_rays(event)
        else:
            self.origins.clear()
            self.rays.clear()

    def detect_rays(self, event: QMouseEvent) -> None:
        x, y, z = [event.pos().x(), event.pos().y(), 1.0]
        self.rays.append(self.viewer.ray_from_click(x, y, z))
        self.origins.append(self.viewer.origin_from_click(x, y, z))

        if len(self.rays) == 2:
            self.viewer.generate_slice_description(self.origins, self.rays)
            self.origins.clear()
            self.rays.clear()



#!/usr/bin/env python


#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from qtpy.QtGui import QMouseEvent, QWheelEvent


class Mode:
    def __init__(self, viewer):
        self.name = 'Base mode'
        self.viewer = viewer

    def mousePressEvent(self, event: QMouseEvent) -> None:
        self.viewer.signal_screen_clicked.emit([event.x(), event.y(), 0])

    def mouseMoveEvent(self, event: QMouseEvent) -> None:
        pass

    def mouseReleaseEvent(self, event: QMouseEvent) -> None:
        pass

    def mouseDoubleClickEvent(self, event: QMouseEvent) -> None:
        pass

    def wheelEvent(self, event: QWheelEvent) -> None:
        dy = event.angleDelta().y()
        sign = dy / abs(dy) if abs(dy) > 1e-12 else 0.0
        smoothness = max(self.viewer.smoothness, 0.1)

        # Arbitrary number, but dependent on viewer's smoothness
        rate = 0.4 / smoothness
        movement_rate = rate / (1.0 + max(0.0, min(sign, rate)))

        off_center = self.viewer.off_center[2]
        shift = -sign * movement_rate * max(-sign, off_center) / smoothness

        self.viewer.translate(x=0, y=0, z=shift)



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from .glcollection import GLCollection

from ..drawables.meshgl import MeshGL
from ..drawables.blockgl import BlockGL
from ..drawables.linegl import LineGL
from ..drawables.pointgl import PointGL
from ..drawables.tubegl import TubeGL

from ..glprograms.meshprogram import MeshProgram
from ..glprograms.wireprogram import WireProgram
from ..glprograms.highlightprogram import HighlightProgram
from ..glprograms.turbomeshprogram import TurboMeshProgram
from ..glprograms.crosssectionprogram import CrossSectionProgram

from ..glprograms.blockprogram import BlockProgram
from ..glprograms.blocklegacyprogram import BlockLegacyProgram

from ..glprograms.lineprogram import LineProgram
from ..glprograms.pointprogram import PointProgram
from ..glprograms.tubeprogram import TubeProgram


class GLDrawableCollection(GLCollection):
    def __init__(self, viewer=None):
        super().__init__(viewer)
        self.associate(LineProgram(viewer), lambda: self.retrieve(LineGL))
        self.associate(TubeProgram(viewer), lambda: self.retrieve(TubeGL))
        self.associate(PointProgram(viewer), lambda: self.retrieve(PointGL))

        # Blocks
        self.associate(BlockLegacyProgram(viewer), lambda: self.retrieve(BlockGL, 'block_legacy'))
        self.associate(BlockProgram(viewer), lambda: self.retrieve(BlockGL, 'block_standard'))

        # Meshes
        self.associate(MeshProgram(viewer), lambda: self.retrieve(MeshGL, 'mesh_standard'))
        self.associate(WireProgram(viewer), lambda: self.retrieve(MeshGL, 'mesh_wireframe'))
        self.associate(HighlightProgram(viewer), lambda: self.retrieve(MeshGL, 'mesh_highlight'))
        self.associate(TurboMeshProgram(viewer), lambda: self.retrieve(MeshGL, 'mesh_turbo'))
        self.associate(CrossSectionProgram(viewer), lambda: self.retrieve(MeshGL, 'mesh_cross'))



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import numpy as np
from OpenGL.GL import *

from .meshprogram import MeshProgram
from ..drawables.gldrawable import GLDrawable


class TurboMeshProgram(MeshProgram):
    def __init__(self, viewer):
        super().__init__(viewer)
        self.base_name = 'MeshTurbo'
        self.info = {
            'opaque': {
                'vaos': [],
                'vbos': [],
                'num_indices': 0,
            },
            'transparent': {
                'vaos': [],
                'vbos': [],
                'num_indices': 0,
            },
        }

    def set_drawables(self, drawables: list) -> None:
        super().set_drawables(drawables)
        self.set_buffers(self.opaques, 'opaque')
        self.set_buffers(self.transparents, 'transparent')

    def set_buffers(self, meshes: list, visibility: str) -> None:
        _POSITION = 0
        _COLOR = 1

        # VBO
        if len(self.info[visibility]['vbos']) == 0:
            self.info[visibility]['vaos'] = [glGenVertexArrays(1)]
            self.info[visibility]['vbos'] = glGenBuffers(3)

        # Data
        vertices = np.empty(len(meshes), np.ndarray)
        indices = np.empty(len(meshes), np.ndarray)
        colors = np.empty(len(meshes), np.ndarray)

        vertices_counter = 0
        indices_counter = 0

        for index, mesh in enumerate(meshes):
            num_vertices = len(mesh.element.vertices)

            vertices[index] = mesh.element.vertices.astype(np.float32)
            indices[index] = (mesh.element.indices + vertices_counter)
            colors[index] = np.tile(mesh.element.rgba, num_vertices).astype(np.float32)

            vertices_counter += num_vertices
            indices_counter += mesh.element.indices.size

        self.info[visibility]['num_indices'] = indices_counter

        if len(meshes) > 0:
            vertices = np.concatenate(vertices)
            indices = np.concatenate(indices)
            colors = np.concatenate(colors)

        glBindVertexArray(self.info[visibility]['vaos'][-1])

        # Recycle GLDrawable's fill_buffers method
        GLDrawable.fill_buffer(_POSITION, 3, vertices, GLfloat, GL_FLOAT,
                               self.info[visibility]['vbos'][_POSITION])
        GLDrawable.fill_buffer(_COLOR, 4, colors, GLfloat, GL_FLOAT,
                               self.info[visibility]['vbos'][_COLOR])

        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self.info[visibility]['vbos'][-1])
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, indices, GL_STATIC_DRAW)

        glBindVertexArray(0)

    def draw(self) -> None:
        glBindVertexArray(self.info['opaque']['vaos'][-1])
        glDrawElements(GL_TRIANGLES, self.info['opaque']['num_indices'], GL_UNSIGNED_INT, None)
        glBindVertexArray(0)

    def redraw(self) -> None:
        glBindVertexArray(self.info['transparent']['vaos'][-1])
        glDepthMask(GL_FALSE)
        glEnable(GL_CULL_FACE)

        for gl_cull in [GL_FRONT, GL_BACK]:
            glCullFace(gl_cull)
            glDrawElements(GL_TRIANGLES, self.info['transparent']['num_indices'], GL_UNSIGNED_INT, None)

        glDisable(GL_CULL_FACE)
        glDepthMask(GL_TRUE)
        glBindVertexArray(0)



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import h5py
import numpy as np
from qtpy.QtCore import QFileInfo
from .parserdata import ParserData
from .parser import Parser


class H5MParser(Parser):
    @staticmethod
    def load_file(path: str, *args, **kwargs) -> ParserData:
        hf = h5py.File(path, 'r')
        vertices = hf['vertices'][()]
        indices = hf['indices'][()]
        properties = dict(hf.attrs)
        hf.close()

        # Metadata
        properties['name'] = QFileInfo(path).completeBaseName()
        properties['extension'] = QFileInfo(path).suffix()

        data = ParserData()
        data.vertices = vertices
        data.indices = indices
        data.properties = properties

        return data

    @staticmethod
    def save_file(*args, **kwargs) -> None:
        path = kwargs.get('path')

        if path is None:
            raise KeyError('Path missing.')

        vertices = kwargs.get('vertices')
        indices = kwargs.get('indices')

        if vertices is None:
            data = kwargs.get('data', {})
            try:
                vertices = data['vertices']
            except Exception:
                vertices = np.column_stack((data.get('x', []),
                                            data.get('y', []),
                                            data.get('z', []),
                                            ))

            indices = indices or data.get('indices', [])

        properties = kwargs.get('properties', {})

        hf = h5py.File(path, 'w')
        hf.create_dataset('vertices', data=vertices, dtype='float32')
        hf.create_dataset('indices', data=indices, dtype='uint32')

        for k, v in properties.items():
            hf.attrs[k] = v

        hf.close()



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import numpy as np
import pandas as pd

from qtpy.QtCore import QFileInfo
from .parserdata import ParserData
from .parser import Parser


class OFFParser(Parser):
    @staticmethod
    def load_file(path: str, *args, **kwargs) -> ParserData:
        assert path.lower().endswith('off')

        with open(path, 'r') as fp:
            assert 'OFF' == fp.readline().strip()
            n_vertices, n_faces, n_edges = tuple([int(s) for s in fp.readline().strip().split(' ')])

            vertices = pd.read_csv(path, skiprows=1, nrows=n_vertices,
                                   usecols=range(3), sep=' ').to_numpy(dtype=float)
            indices = pd.read_csv(path, skiprows=1 + n_vertices, nrows=n_faces, delimiter=' ',
                                  usecols=range(1, 4)).to_numpy(dtype=int)

            # Metadata
            properties = {
                'name': QFileInfo(path).completeBaseName(),
                'extension': QFileInfo(path).suffix()
            }

            data = ParserData()
            data.vertices = vertices
            data.indices = indices
            data.properties = properties

            return data

    @staticmethod
    def save_file(path: str, *args, **kwargs) -> None:
        if path is None:
            raise KeyError('Path missing.')

        vertices = kwargs.get('vertices')
        indices = kwargs.get('indices')

        if vertices is None:
            data = kwargs.get('data', {})
            if 'vertices' in data.keys():
                vertices = data['vertices']
            else:
                vertices = np.column_stack((data.get('x', []),
                                            data.get('y', []),
                                            data.get('z', []),
                                            ))

            indices = indices or data.get('indices', [])

        V = len(vertices)
        F = len(indices)
        E = V + F - 2

        # Internal data
        with open(path, 'w') as f:
            f.write('OFF\n')
            f.write(f'{V} {F} {E}\n')

        # Vertices
        pd.DataFrame(vertices).to_csv(path, mode='a', sep=' ', header=False, index=False)

        # Indices
        pd.DataFrame(np.column_stack(([[3]] * len(indices), indices))).to_csv(
            path, mode='a', sep=' ', header=False, index=False)



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import numpy as np

from colour import Color


def magnitude(vector: np.ndarray) -> float:
    return np.linalg.norm(vector)


def normalize(vector: np.ndarray) -> np.ndarray:
    return vector / magnitude(vector)


def dot_by_row(vectors_a: np.ndarray, vectors_b: np.ndarray) -> np.ndarray:
    # Returns the row-wise dot product of the input
    # Ex.: dot_by_row([v1, v2, v3], [v4, v5, v6]) = [dot(v1, v4), dot(v2, v5), dot(v3, v6)]
    return np.einsum('ij,ij->i', vectors_a, vectors_b)


def magnitude2_by_row(vectors: np.ndarray) -> np.ndarray:
    # Returns the squared magnitudes of each vector in the input
    # Ex.: magnitude2([v1, v2]) = [length(v1)^2, length(v)^2],
    # where length is the magnitude of a single vector
    return dot_by_row(vectors, vectors)


def magnitude_by_row(vectors: np.ndarray) -> np.ndarray:
    # Returns the magnitudes of each vector in the input
    # Ex.: magnitude([v1, v2]) = [length(v1), length(v2)]
    return np.sqrt(magnitude2_by_row(vectors))


def closest_point_to(origin: np.ndarray, points: np.ndarray) -> np.ndarray or None:
    if points.size == 0:
        return None

    distances = distances_between(origin, points)
    mask = np.abs(distances - distances.min()) <= 1e-12

    return points[mask][0]


def distances_between(origin: np.ndarray, points: np.ndarray) -> np.ndarray or None:
    if points.size == 0:
        return None

    return np.linalg.norm(origin - points, axis=1)


def aabb_intersection(origin: np.ndarray,
                      ray: np.ndarray,
                      b_min: np.ndarray,
                      b_max: np.ndarray) -> bool:
    # Adapted from https://tavianator.com/fast-branchless-raybounding-box-intersections-part-2-nans/
    if (b_max - b_min).min() < 1e-12:  # Flat mesh means AABB unreliable
        return True

    # Result of division by zero used deliberately
    with np.errstate(divide='ignore'):
        ray_inv = 1.0 / ray

    tmin = -np.inf
    tmax = +np.inf

    for i in range(3):
        t1 = (b_min[i] - origin[i]) * ray_inv[i]
        t2 = (b_max[i] - origin[i]) * ray_inv[i]

        tmin = max(tmin, min(t1, t2))
        tmax = min(tmax, max(t1, t2))

    return tmax > max(tmin, 0.0)


def plane_intersection(origin: np.ndarray,
                       ray: np.ndarray,
                       plane_normal: np.ndarray,
                       plane_d: np.ndarray) -> np.ndarray:
    # Taken from https://courses.cs.washington.edu/courses/cse457/09au/lectures/triangle_intersection.pdf
    t = (plane_d - np.dot(plane_normal, origin)) / np.dot(plane_normal, ray)

    return origin + t * ray


def points_inside_mesh(mesh, point_vertices: np.ndarray) -> np.ndarray:
    # With ray tracing, we'll detect which points are inside the mesh
    ray = np.array([0.0, 0.0, 1.0])  # Arbitrary direction
    mask = []

    # From the point center, if we hit the mesh an odd number of times, we're inside the mesh
    for origin in point_vertices:
        intersections = mesh.intersect_with_ray(origin, ray)
        mask.append(len(intersections) > 0 and len(np.unique(intersections, axis=0)) % 2 == 1)

    return np.array(mask)


def mineral_density(mesh, blocks, mineral: str) -> tuple:
    """
    Receives a MeshElement, a BlockElement, a mineral string,
    and returns a tuple consisting of:
    (blocks inside the mesh, values of those blocks, sum of those values)

    :param mesh: MeshElement
    :param blocks: BlockElement
    :param mineral: str
    :return: tuple(np.ndarray, np.ndarray, float)
    """
    # Limit points by new bounding box of mesh and get their values
    min_bound, max_bound = mesh.bounding_box
    block_size = blocks.block_size
    half_block = block_size / 2

    # Recreate bounds to allow a block to be partially inside a mesh
    min_bound -= block_size
    max_bound += block_size

    mask = np.all((blocks.vertices > min_bound) & (blocks.vertices < max_bound), axis=-1)
    bound_blocks = blocks.vertices[mask]
    bound_values = blocks.data.get(mineral)[mask]

    # Get points inside mesh
    mask = points_inside_mesh(mesh, bound_blocks)
    insiders = np.ones(mask.size)

    # We have to check the 8 vertices to know we're not omitting a block just because its
    # center is not inside the mesh.
    # Maybe there's a better way, but we still need to investigate that.
    bias_list = [[-1, -1, -1],
                 [+1, -1, -1],
                 [-1, +1, -1],
                 [+1, +1, -1],
                 [-1, -1, +1],
                 [+1, -1, +1],
                 [-1, +1, +1],
                 [+1, +1, +1]]

    for bias in bias_list:
        mask = mask | points_inside_mesh(mesh, bound_blocks + half_block * bias)
        insiders = mask & points_inside_mesh(mesh, bound_blocks + half_block * bias)

    # TODO Calculate mineral density of blocks partially inside the mesh
    # Until now, we can only detect which blocks are fully/partially/not inside the mesh
    mask_partials = insiders ^ mask  # ^ = XOR
    # mask = mask_partials

    return bound_blocks[mask], bound_values[mask], bound_values[mask].sum()


def values_to_rgb(values: np.ndarray, vmin: float, vmax: float, colormap: str) -> np.ndarray:
    initial, final = parse_colormap(colormap)
    vals = np.interp(np.clip(values, vmin, vmax), (vmin, vmax), (0.0, 1.0))

    hsv = np.ones((vals.size, 3))
    hsv[:, 0] = initial[0] + (final - initial)[0] * vals
    hsv[:, 1] = initial[1] + (final - initial)[1] * vals
    hsv[:, 2] = initial[2] + (final - initial)[2] * vals

    return hsv_to_rgb(hsv)


def parse_colormap(colormap: str) -> list:
    try:
        initial_str, final_str = colormap.split('-')
        initial = np.array(hsl_to_hsv(*Color(initial_str).get_hsl()))
        final = np.array(hsl_to_hsv(*Color(final_str).get_hsl()))
        return [initial, final]
    except Exception:
        return []


def hsl_to_hsv(h: float, s: float, l: float) -> tuple:
    # Taken and adapted from https://gist.github.com/mathebox/e0805f72e7db3269ec22
    v = (2 * l + s * (1 - abs(2 * l - 1))) / 2
    s = 2 * (v - l) / max(v, 1e-12)
    return h, s, v


def hsv_to_rgb(hsv: np.ndarray or list) -> np.ndarray:
    # Taken and adapted from matplotlib.colors
    hsv = np.asarray(hsv)

    h = hsv[:, 0]
    s = hsv[:, 1]
    v = hsv[:, 2]

    r = np.empty_like(h)
    g = np.empty_like(h)
    b = np.empty_like(h)

    i = (h * 6.0).astype(int)
    f = (h * 6.0) - i
    p = v * (1.0 - s)
    q = v * (1.0 - s * f)
    t = v * (1.0 - s * (1.0 - f))

    idx = i % 6 == 0
    r[idx] = v[idx]
    g[idx] = t[idx]
    b[idx] = p[idx]

    idx = i == 1
    r[idx] = q[idx]
    g[idx] = v[idx]
    b[idx] = p[idx]

    idx = i == 2
    r[idx] = p[idx]
    g[idx] = v[idx]
    b[idx] = t[idx]

    idx = i == 3
    r[idx] = p[idx]
    g[idx] = q[idx]
    b[idx] = v[idx]

    idx = i == 4
    r[idx] = t[idx]
    g[idx] = p[idx]
    b[idx] = v[idx]

    idx = i == 5
    r[idx] = v[idx]
    g[idx] = p[idx]
    b[idx] = q[idx]

    idx = s == 0
    r[idx] = v[idx]
    g[idx] = v[idx]
    b[idx] = v[idx]

    rgb = np.stack([r, g, b], axis=-1)

    return rgb



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

import numpy as np

from qtpy.QtCore import Qt
from qtpy.QtWidgets import *


class PropertiesDialog(QDialog):
    def __init__(self, element, parent=None):
        QDialog.__init__(self, parent)
        self.resize(450, 500)

        self.setWindowTitle(f'Set properties ({element.name}.{element.extension})')
        self.setWindowFlags(Qt.WindowStaysOnTopHint)

        # Left side of the dialog
        self.label_x = QLabel('Coordinate X (Easting)', self)
        self.label_y = QLabel('Coordinate Y (Northing)', self)
        self.label_z = QLabel('Coordinate Z (Elevation)', self)
        self.label_values = QLabel('Current value', self)
        self.label_properties = QLabel('Properties', self)

        # Right side of the dialog
        self.comboBox_x = QComboBox(self)
        self.comboBox_y = QComboBox(self)
        self.comboBox_z = QComboBox(self)
        self.comboBox_values = QComboBox(self)
        self.tableWidget_properties = QTableWidget(self)
        self.tableWidget_properties.setColumnCount(1)
        self.tableWidget_properties.setRowCount(0)
        self.tableWidget_properties.setHorizontalHeaderItem(0, QTableWidgetItem('Value'))
        self.tableWidget_properties.horizontalHeader().setStretchLastSection(True)

        # Accept/cancel buttons
        self.buttonBox = QDialogButtonBox(self)
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)

        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

        # Layout
        self.layout = QGridLayout(self)
        self.append_to_grid(0, self.label_x, self.comboBox_x)
        self.append_to_grid(1, self.label_y, self.comboBox_y)
        self.append_to_grid(2, self.label_z, self.comboBox_z)
        self.append_to_grid(3, self.label_values, self.comboBox_values)
        self.append_to_grid(4, self.label_properties, self.tableWidget_properties)
        self.append_to_grid(5, QWidget(), self.buttonBox)

        # Fill headers and set current ones
        self.fill_headers(element.all_headers)
        self.current_headers = element.headers

        # Fill properties in QTableWidget
        self.tableWidget_properties.setRowCount(len(element.customizable_properties))
        self.tableWidget_properties.setVerticalHeaderLabels(element.customizable_properties)

        for i, k in enumerate(element.customizable_properties):
            v = getattr(element, k)
            text = str(v.tolist()) if type(v) is np.ndarray else str(v)
            self.tableWidget_properties.setItem(i, 0, QTableWidgetItem(text))

    def append_to_grid(self, row: int, left: QWidget, right: QWidget) -> None:
        self.layout.addWidget(left, row, 0, 1, 1)
        self.layout.addWidget(right, row, 1, 1, 1)

    def fill_headers(self, headers: list) -> None:
        for i in headers:
            self.comboBox_x.addItem(i)
            self.comboBox_y.addItem(i)
            self.comboBox_z.addItem(i)
            self.comboBox_values.addItem(i)

    def has_altered_coordinates(self, element) -> bool:
        altered = False
        for e, c in list(zip(element.headers, self.current_headers))[:3]:
            altered |= not (e == c)

        return altered

    @property
    def current_headers(self) -> list:
        return [self.comboBox_x.currentText(),
                self.comboBox_y.currentText(),
                self.comboBox_z.currentText(),
                self.comboBox_values.currentText()]

    @current_headers.setter
    def current_headers(self, headers: list) -> None:
        self.comboBox_x.setCurrentIndex(self.comboBox_x.findText(headers[0]))
        self.comboBox_y.setCurrentIndex(self.comboBox_y.findText(headers[1]))
        self.comboBox_z.setCurrentIndex(self.comboBox_z.findText(headers[2]))
        self.comboBox_values.setCurrentIndex(self.comboBox_values.findText(headers[3]))



#!/usr/bin/env python

#  Copyright (c) 2019-2020 Gabriel Sanhueza.
#
#  Distributed under the MIT License.
#  See LICENSE for more info.

from datetime import datetime

from qtpy.QtWidgets import QFileDialog
from qtpy.QtWidgets import QVBoxLayout
from qtpy.QtWidgets import QWidget

from .camerawidget import CameraWidget
from .toolbar import ToolBar
from .treewidget import TreeWidget
from .iconcollection import IconCollection
from ..integrableviewer import IntegrableViewer


class Container(QWidget):
    def __init__(self, parent=None):
        QWidget.__init__(self, parent)
        self.setAcceptDrops(True)

        # Container
        self.setWindowTitle('BlastSight (Container)')
        self.resize(600, 500)

        # Widgets
        self.toolbar = ToolBar(self)
        self.viewer = IntegrableViewer(self)
        self.camera = CameraWidget()
        self.tree = TreeWidget()

        # Layout
        self.layout = QVBoxLayout(self)
        self.layout.addWidget(self.viewer)
        self.layout.addWidget(self.toolbar)

        # Actions
        self.toolbar.auto_connect(self.tree, self.viewer, self.camera)

        actions = self.toolbar.action_collection
        actions.action_take_screenshot.triggered.connect(self.handle_screenshot)

        # Icons
        try:
            self.set_widgets_icon(parent.windowIcon())
        except AttributeError:
            self.set_widgets_icon(IconCollection.get('blastsight.png'))

    def set_widgets_icon(self, icon) -> None:
        self.setWindowIcon(icon)
        self.camera.setWindowIcon(icon)
        self.tree.setWindowIcon(icon)

    def handle_screenshot(self) -> None:
        (path, selected_filter) = QFileDialog.getSaveFileName(
            parent=self.viewer,
            directory=f'BlastSight Screenshot ({datetime.now().strftime("%Y%m%d-%H%M%S")})',
            filter='PNG image (*.png);;')

        self.viewer.take_screenshot(path)

    def dragEnterEvent(self, event, *args, **kwargs) -> None:
        self.viewer.dragEnterEvent(event, *args, **kwargs)

    def dropEvent(self, event, *args, **kwargs) -> None:
        self.viewer.dropEvent(event, *args, **kwargs)
